#include "eanim.h"

void decode_pixel(int code, Pixel24 *p, int pix);

Display *disp;
Window win;
Visual *vis;
int scr;
GC gc;
int depth;
int winwidth;
int winheight;
int imgjmp;
char shmp;
Colormap cmap;
XImage *xim;
XShmSegmentInfo shminfo;
Pixmap pmap;


void qit()
{
   printf("\n\n\nClean Exit\n\n\n");
   XShmDetach(disp,&shminfo);
   shmdt(shminfo.shmaddr);
   shmctl(shminfo.shmid,IPC_RMID,0);
   XDestroyImage(xim);
   XFreePixmap(disp,pmap);
   XSetWindowBackgroundPixmap(disp,win,0);
   XSetWindowBackground(disp,win,0);
   XClearWindow(disp,win);
   exit(0);
}

void X_SetupImage(int w,int h)
{
   xim=XShmCreateImage(disp,vis,depth,ZPixmap,NULL,&shminfo,w,h);
   if (!xim)
     fprintf(stderr,"X-Error: unable to create XShm XImage\n");
   shminfo.shmid=shmget(IPC_PRIVATE,
			xim->bytes_per_line*xim->height,
			IPC_CREAT|0777);
   if (shminfo.shmid==-1)
     fprintf(stderr,"SharedMemory Error: unable to get identifier\n");
   shminfo.shmaddr=xim->data=shmat(shminfo.shmid,0,0);
   if(!XShmAttach(disp,&shminfo))
     fprintf(stderr,"X-Error: unable to attach XShm Shared Memory Segment\n");
   if (shmp)
     pmap=XShmCreatePixmap(disp,win,shminfo.shmaddr,&shminfo,w,h,depth);
   else
     pmap=XCreatePixmap(disp,win,w,h,depth);
   if (!pmap)
     fprintf(stderr,"Unable to create pixmap\n");
   imgjmp=w;
   XSetWindowBackgroundPixmap(disp,win,pmap);
}

int X_DisplayImage(char *block, int bx, int by, int glob, int csize)
{
   int x,y,xx,yy,gx,gy;
   int show,i,cnt,shift;
   
   if      (csize==8)   shift=3;
   else if (csize==16)  shift=4;
   else if (csize==32)  shift=5;
   else if (csize==64)  shift=6;
   else if (csize==128) shift=7;
   cnt=0;
   if (block)
     {
	for (y=0;y<by;y+=glob)
	  {
	     for (x=0;x<bx;x+=glob)
	       {
		  show=0;
		  if ((bx-x)<glob) gx=bx-x;
		  else gx=glob;
		  if ((by-y)<glob) gy=by-y;
		  else gy=glob;
		  for (yy=0;yy<gy;yy++)
		    {
		       i=((y+yy)*bx)+x;
		       for (xx=0;xx<gx;xx++) show+=block[i++];
		    }
		  if (show)
		    {
		       if (!shmp)
			 XShmPutImage(disp,pmap,gc,xim,x<<shift,y<<shift,
				      x<<shift,y<<shift,gx<<shift,gy<<shift,
				      False);
		       XClearArea(disp,win,x<<shift,y<<shift,gx<<shift,
				  gy<<shift,False);
		       cnt+=glob*glob;
		    }
	       }
	  }
     }
   else
     {
	if (!shmp)
	  XShmPutImage(disp,pmap,gc,xim,0,0,0,0,winwidth,winheight,False);
	XClearWindow(disp,win);
     }
   return cnt;
}

void X_OpenWindow(int x, int y, int w,int h, int rr)
{
   XGCValues gcvals;
   Window root;
   XColor xcl;
   int i;
   int r,g,b;
   
   disp=XOpenDisplay(NULL);
   if (disp)
     {
	scr=DefaultScreen(disp);
	vis=DefaultVisual(disp,scr);
	root=DefaultRootWindow(disp);
	depth=DefaultDepth(disp,scr);
	if (!rr) win=XCreateSimpleWindow(disp,root,x,y,w,h,0,0,0);
	else win=root;
	gc=XCreateGC(disp,win,(unsigned long)0,&gcvals);
	XSetForeground(disp,gc,0);
	XSetBackground(disp,gc,1);
	if (!rr) XMapWindow(disp,win);
	XFlush(disp);
	winwidth=w;
	winheight=h;
     }
}

void X_Sync()
{
   XSync(disp,False);
}

void X_Dechunkify(int w, int h, int x, int y, void *c, int csize)
{
   int xx,yy,x2,y2,i,j,val,shift;
   RawChunk24_8   *c8;
   RawChunk24_16  *c16;
   RawChunk24_32  *c32;
   RawChunk24_64  *c64;
   RawChunk24_128 *c128;
   
   if      (csize==8)   c8=(RawChunk24_8 *)c;
   else if (csize==16)  c16=(RawChunk24_16 *)c;
   else if (csize==32)  c32=(RawChunk24_32 *)c;
   else if (csize==64)  c64=(RawChunk24_64 *)c;
   else if (csize==128) c128=(RawChunk24_128 *)c;
   if      (csize==8)   shift=3;
   else if (csize==16)  shift=4;
   else if (csize==32)  shift=5;
   else if (csize==64)  shift=6;
   else if (csize==128) shift=7;
   i=0;
   if (csize==8)
     {
	for(yy=0;yy<y;yy++)
	  {
	     for(xx=0;xx<x;xx++)
	       {
		  for(y2=0;y2<csize;y2++)
		    {
		       for(x2=0;x2<csize;x2++)
			 {
			    val=(c8[i].p[x2][y2].r<<8)&0xf800|
			      (c8[i].p[x2][y2].g<<3)&0x07e0|
			      (c8[i].p[x2][y2].b>>3)&0x001f;
			    XPutPixel(xim,(xx<<shift)+x2,(yy<<shift)+y2,val);
			 }
		    }
		  i++;
	       }
	  }
     }
   else if (csize==16)
     {
	for(yy=0;yy<y;yy++)
	  {
	     for(xx=0;xx<x;xx++)
	       {
		  for(y2=0;y2<csize;y2++)
		    {
		       for(x2=0;x2<csize;x2++)
			 {
			    val=(c16[i].p[x2][y2].r<<8)&0xf800|
			      (c16[i].p[x2][y2].g<<3)&0x07e0|
			      (c16[i].p[x2][y2].b>>3)&0x001f;
			    XPutPixel(xim,(xx<<shift)+x2,(yy<<shift)+y2,val);
			 }
		    }
		  i++;
	       }
	  }
     }
   else if (csize==32)
     {
	for(yy=0;yy<y;yy++)
	  {
	     for(xx=0;xx<x;xx++)
	       {
		  for(y2=0;y2<csize;y2++)
		    {
		       for(x2=0;x2<csize;x2++)
			 {
			    val=(c32[i].p[x2][y2].r<<8)&0xf800|
			      (c32[i].p[x2][y2].g<<3)&0x07e0|
			      (c32[i].p[x2][y2].b>>3)&0x001f;
			    XPutPixel(xim,(xx<<shift)+x2,(yy<<shift)+y2,val);
			 }
		    }
		  i++;
	       }
	  }
     }
   else if (csize==64)
     {
	for(yy=0;yy<y;yy++)
	  {
	     for(xx=0;xx<x;xx++)
	       {
		  for(y2=0;y2<csize;y2++)
		    {
		       for(x2=0;x2<csize;x2++)
			 {
			    val=(c64[i].p[x2][y2].r<<8)&0xf800|
			      (c64[i].p[x2][y2].g<<3)&0x07e0|
			      (c64[i].p[x2][y2].b>>3)&0x001f;
			    XPutPixel(xim,(xx<<shift)+x2,(yy<<shift)+y2,val);
			 }
		    }
		  i++;
	       }
	  }
     }
   else if (csize==128)
     {
	for(yy=0;yy<y;yy++)
	  {
	     for(xx=0;xx<x;xx++)
	       {
		  for(y2=0;y2<csize;y2++)
		    {
		       for(x2=0;x2<csize;x2++)
			 {
			    val=(c128[i].p[x2][y2].r<<8)&0xf800|
			      (c128[i].p[x2][y2].g<<3)&0x07e0|
			      (c128[i].p[x2][y2].b>>3)&0x001f;
			    XPutPixel(xim,(xx<<shift)+x2,(yy<<shift)+y2,val);
			 }
		    }
		  i++;
	       }
	  }
     }
   X_DisplayImage(NULL,0,0,0,csize);
}

void X_Decode_chunk3(int code_type, void *data, int x2, int y2, int csize)
{
   Pixel24 p;
   int bpp,x,y,xx,yy,res,i,code,val;
   int x1,y1,xx1,yy1,mul,sz,shift;
   unsigned int pix;
   unsigned char *byte;
   unsigned short *word;

   res=code_type&FLAG_RES_MSK;
   code=code_type&FLAG_PIX_MSK;
   if (code==FLAG_PIX_24b) bpp=3;
   else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
	    (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (bpp==2) word=(unsigned short *)data;
   else byte=(unsigned char *)data;
   i=0;
   if      (res==FLAG_RES_1x1)      sz=1;
   else if (res==FLAG_RES_2x2)      sz=2;
   else if (res==FLAG_RES_4x4)      sz=4;
   else if (res==FLAG_RES_8x8)      sz=8;
   else if (res==FLAG_RES_16x16)    sz=16;
   else if (res==FLAG_RES_32x32)    sz=32;
   else if (res==FLAG_RES_64x64)    sz=64;
   else if (res==FLAG_RES_128x128)  sz=128;
   if      (csize==8)   shift=3;
   else if (csize==16)  shift=4;
   else if (csize==32)  shift=5;
   else if (csize==64)  shift=6;
   else if (csize==128) shift=7;
   mul=csize/sz;
/* defines for evaluating pixel valuse for 16bpp 565 */
#define VAL_24b p.r=byte[i++];p.g=byte[i++];p.b=byte[i++];val=((p.r<<8)&0xf800)|((p.g<<3)&0x07e0)|((p.b>>3)&0x001f)
#define VAL_16b val=(word[i++])
#define VAL_RG  val=((word[i])&0xf800)|((word[i++]<<3)&0x07e0)
#define VAL_RB  val=((word[i])&0xf800)|((word[i++]>>3)&0x001f)
#define VAL_GB  val=((word[i]>>5)&0x07e0)|((word[i++]>>3)&0x001f)
#define VAL_332 val=((byte[i]<<8)&0xe000)|((byte[i]<<6)&0x0700)|((byte[i++]<<3)&0x0018)
#define VAL_R   val=((byte[i++]<<8)&0xf800)
#define VAL_G   val=((byte[i++]<<3)&0x07e0)
#define VAL_B   val=((byte[i++]>>3)&0x001f)
#define VAL_RG8 val=((byte[i]<<8)&0xf000)|((byte[i++]<<7)&0x0780)
#define VAL_RB8 val=((byte[i]<<8)&0xf000)|((byte[i++]<<1)&0x001e)
#define VAL_GB8 val=((byte[i]<<3)&0x0780)|((byte[i++]<<1)&0x001e)
#define VAL_GRY val=((byte[i]<<8)&0xf800)|((byte[i]<<3)&0x07e0)|((byte[i++]>>3)&0x001f)
#define VAL_422 val=((byte[i]<<8)&0xf000)|((byte[i]<<7)&0x0600)|((byte[i++]<<3)&0x0018)
#define VAL_242 val=((byte[i]<<8)&0xc000)|((byte[i]<<5)&0x0780)|((byte[i++]<<3)&0x0018)
#define VAL_224 val=((byte[i]<<8)&0xc000)|((byte[i]<<5)&0x0600)|((byte[i++]<<1)&0x001e)
   x1=x2<<shift;y1=y2<<shift;
   xx1=x1+csize;yy1=y1+csize;
   switch (code)
     {
      case FLAG_PIX_24b:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_24b;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_16b:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_16b;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_RG:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_RG;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_RB:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_RB;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_GB:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_GB;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_332:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_332;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_R:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_R;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_G:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_G;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_B:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_B;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_RG8:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_RG8;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_RB8:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_RB8;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_GB8:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_GB8;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_GRY:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_GRY;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_422:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_422;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_242:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_242;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      case FLAG_PIX_224:
	for (y=y1;y<yy1;y+=mul)
	  {
	     for(x=x1;x<xx1;x+=mul)
	       {
		  VAL_224;
		  for(yy=0;yy<mul;yy++)
		    {
		       for(xx=0;xx<mul;xx++)
			 {
			    XPutPixel(xim,x+xx,y+yy,val);
			 }
		    }
	       }
	  }
	break;
      default:
	break;
     }
}

int X_Load_chunks_frame_m(unsigned char *m, int *size, int *w, int *h, int *x, int *y, int csize)
{
   int i,code,bpp,cd,sz,xx,yy;
   void *data;
   int *ip;
   unsigned char *ptr;
   
   ptr=m;
   ip=(int *)ptr;
   code=*ip;ptr+=sizeof(int);
   if (code==0xffffffff) 
     {
	*size=ptr-m;
	return 0;
     }
   else if (code==0xfffffffe) 
     {
	*size=ptr-m;
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if      (sz==FLAG_RES_128x128) bpp*=128*128;
   else if (sz==FLAG_RES_64x64)   bpp*=64*64;
   else if (sz==FLAG_RES_32x32)   bpp*=32*32;
   else if (sz==FLAG_RES_16x16)   bpp*=16*16;
   else if (sz==FLAG_RES_8x8)     bpp*=8*8;
   else if (sz==FLAG_RES_4x4)     bpp*=4*4;
   else if (sz==FLAG_RES_2x2)     bpp*=2*2;
   data=(void *)ptr;
   ptr+=bpp;
   i=(code>>7)&0x01ffffff;
   xx=i%*x;
   yy=(i/(*x));
   X_Decode_chunk3(code,data,xx,yy,csize);
   *size=ptr-m;
   return i+1;
}

int X_Skip_chunks_frame_m(unsigned char *m, int *size)
{
   int i,code,bpp,cd,sz,xx,yy;
   void *data;
   int *ip;
   unsigned char *ptr;
   
   ptr=m;
   ip=(int *)ptr;
   code=*ip;ptr+=sizeof(int);
   if (code==0xffffffff) 
     {
	*size=ptr-m;
	return 0;
     }
   else if (code==0xfffffffe) 
     {
	*size=ptr-m;
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if      (sz==FLAG_RES_128x128) bpp*=128*128;
   else if (sz==FLAG_RES_64x64)   bpp*=64*64;
   else if (sz==FLAG_RES_32x32)   bpp*=32*32;
   else if (sz==FLAG_RES_16x16)   bpp*=16*16;
   else if (sz==FLAG_RES_8x8)     bpp*=8*8;
   else if (sz==FLAG_RES_4x4)     bpp*=4*4;
   else if (sz==FLAG_RES_2x2)     bpp*=2*2;
   ptr+=bpp;
   *size=ptr-m;
   i=(code>>7)&0x01ffffff;
   return i+1;
}

int X_Skip_chunks_frame_f(FILE *f, int *size)
{
   int i,code,bpp,cd,sz,xx,yy;
   void *data;
   int *ip;
   unsigned char *ptr;
   
   if (!fread(&code,sizeof(unsigned int),1,f))
     {
	return -1;
     }
   if (code==0xffffffff) 
     {
	*size=sizeof(unsigned int);
	return 0;
     }
   else if (code==0xfffffffe) 
     {
	*size=sizeof(unsigned int);
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if      (sz==FLAG_RES_128x128) bpp*=128*128;
   else if (sz==FLAG_RES_64x64)   bpp*=64*64;
   else if (sz==FLAG_RES_32x32)   bpp*=32*32;
   else if (sz==FLAG_RES_16x16)   bpp*=16*16;
   else if (sz==FLAG_RES_8x8)     bpp*=8*8;
   else if (sz==FLAG_RES_4x4)     bpp*=4*4;
   else if (sz==FLAG_RES_2x2)     bpp*=2*2;
   fseek(f,bpp,SEEK_CUR);
   *size=sizeof(unsigned int)+bpp;
   i=(code>>7)&0x01ffffff;
   return i+1;
}

int X_Load_chunks_frame_f(FILE *f, int *w, int *h, int *x, int *y,  int csize)
{
   int i,code,bpp,cd,sz,xx,yy;
   void *data;

   data=(void *)malloc(128*128*3);
   if (!data) return 0;
   if (!fread(&code,sizeof(unsigned int),1,f))
     {
	free(data);
	return -1;
     }
   if (code==0xffffffff) 
     {
	free(data);
	return 0;
     }
   else if (code==0xfffffffe) 
     {
	free(data);
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if      (sz==FLAG_RES_128x128) bpp*=128*128;
   else if (sz==FLAG_RES_64x64)   bpp*=64*64;
   else if (sz==FLAG_RES_32x32)   bpp*=32*32;
   else if (sz==FLAG_RES_16x16)   bpp*=16*16;
   else if (sz==FLAG_RES_8x8)     bpp*=8*8;
   else if (sz==FLAG_RES_4x4)     bpp*=4*4;
   else if (sz==FLAG_RES_2x2)     bpp*=2*2;
   if (!fread(data,bpp,1,f)) 
     {
	free(data);
	fprintf(stderr,"eof\n");
	return -1;
     }
   i=(code>>7)&0x01ffffff;
   xx=i%*x;
   yy=(i/(*x));
   X_Decode_chunk3(code,data,xx,yy,csize);
   free(data);
   return i+1;
}

void decode_pixel(int code, Pixel24 *p, int pix)
{
   unsigned int r,g,b;
   
   switch (code)
     {
      case FLAG_PIX_24b:
	r=(pix>>16)&0xff;
	g=(pix>>8)&0xff;
	b=(pix)&0xff;
	break;
      case FLAG_PIX_16b:
	r=(pix>>8)&0xf8;
	g=(pix>>3)&0xfc;
	b=(pix<<3)&0xf8;
	break;
      case FLAG_PIX_332:
	r=(pix)&0xe0;
	g=(pix<<3)&0xe0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_R:
	r=pix;
	g=0;
	b=0;
	break;
      case FLAG_PIX_G:
	r=0;
	g=pix;
	b=0;
	break;
      case FLAG_PIX_B:
	r=0;
	g=0;
	b=pix;
	break;
      case FLAG_PIX_RG:
	r=(pix>>8);
	g=pix&0xff;
	b=0;
	break;
      case FLAG_PIX_RB:
	r=(pix>>8);
	g=0;
	b=pix&0xff;
	break;
      case FLAG_PIX_GB:
	r=0;
	g=(pix>>8);
	b=pix&0xff;
	break;
      case FLAG_PIX_RG8:
	r=pix&0xf0;
	g=(pix<<4)&0xf0;
	b=0;
	break;
      case FLAG_PIX_RB8:
	r=pix&0xf0;
	g=0;
	b=(pix<<4)&0xf0;
	break;
      case FLAG_PIX_GB8:
	r=0;
	g=pix&0xf0;
	b=(pix<<4)&0xf0;
	break;
      case FLAG_PIX_GRY:
	r=pix;
	g=pix;
	b=pix;
	break;
      case FLAG_PIX_422:
	r=pix&0xf0;
	g=(pix<<4)&0xc0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_242:
	r=pix&0xc0;
	g=(pix<<2)&0xf0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_224:
	r=pix&0xc0;
	g=(pix<<2)&0xc0;
	b=(pix<<4)&0xf0;
	break;
      default:
	break;
     }
   p->r=(unsigned char)r;
   p->g=(unsigned char)g;
   p->b=(unsigned char)b;
}

void decode_chunk(int code_type, void *data, void *chk, int csize)
{
   Pixel24 p;
   int bpp,x,y,xx,yy,res,i,code;
   unsigned int pix;
   unsigned char *byte;
   unsigned short *word;
   int shift; 
   RawChunk24_8   *c8;
   RawChunk24_16  *c16;
   RawChunk24_32  *c32;
   RawChunk24_64  *c64;
   RawChunk24_128 *c128;
   
   if      (csize==8)   c8=(RawChunk24_8 *)chk;
   else if (csize==16)  c16=(RawChunk24_16 *)chk;
   else if (csize==32)  c32=(RawChunk24_32 *)chk;
   else if (csize==64)  c64=(RawChunk24_64 *)chk;
   else if (csize==128) c128=(RawChunk24_128 *)chk;
   if      (csize==8)   shift=3;
   else if (csize==16)  shift=4;
   else if (csize==32)  shift=5;
   else if (csize==64)  shift=6;
   else if (csize==128) shift=7;

   res=code_type&FLAG_RES_MSK;code=code_type&FLAG_PIX_MSK;
   if (code==FLAG_PIX_24b) bpp=3;
   else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
	    (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB)) bpp=2;
   else        bpp=1;
   if (bpp==2) word=(unsigned short *)data;
   else        byte=(unsigned char *)data;
   i=0;
   if (csize==8)
     {
	switch(res)
	  {
	   case FLAG_RES_8x8:
	     for (y=0;y<csize;y++)
	       {
		  for(x=0;x<csize;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);	
		       c8->p[x][y].r=p.r;c8->p[x][y].g=p.g;c8->p[x][y].b=p.b;
		    }
	       }
	     break;
	   case FLAG_RES_4x4:
	     for (y=0;y<csize>>1;y++)
	       {
		  for(x=0;x<csize>>1;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<2;yy++)
			 {
			    for(xx=0;xx<2;xx++)
			      {
				 c8->p[(x<<1)+xx][(y<<1)+yy].r=p.r;
				 c8->p[(x<<1)+xx][(y<<1)+yy].g=p.g;
				 c8->p[(x<<1)+xx][(y<<1)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_2x2:
	     for (y=0;y<csize>>2;y++)
	       {
		  for(x=0;x<csize>>2;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<4;yy++)
			 {
			    for(xx=0;xx<4;xx++)
			      {
				 c8->p[(x<<2)+xx][(y<<2)+yy].r=p.r;
				 c8->p[(x<<2)+xx][(y<<2)+yy].g=p.g;
				 c8->p[(x<<2)+xx][(y<<2)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_1x1:
	     for (y=0;y<csize>>3;y++)
	       {
		  for(x=0;x<csize>>3;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<8;yy++)
			 {
			    for(xx=0;xx<8;xx++)
			      {
				 c8->p[(x<<3)+xx][(y<<3)+yy].r=p.r;
				 c8->p[(x<<3)+xx][(y<<3)+yy].g=p.g;
				 c8->p[(x<<3)+xx][(y<<3)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   default:
	     break;
	  }
     }
   else if (csize==16)
     {
	switch(res)
	  {
	   case FLAG_RES_16x16:
	     for (y=0;y<csize;y++)
	       {
		  for(x=0;x<csize;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);	
		       c16->p[x][y].r=p.r;c16->p[x][y].g=p.g;c16->p[x][y].b=p.b;
		    }
	       }
	     break;
	   case FLAG_RES_8x8:
	     for (y=0;y<csize>>1;y++)
	       {
		  for(x=0;x<csize>>1;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<2;yy++)
			 {
			    for(xx=0;xx<2;xx++)
			      {
				 c16->p[(x<<1)+xx][(y<<1)+yy].r=p.r;
				 c16->p[(x<<1)+xx][(y<<1)+yy].g=p.g;
				 c16->p[(x<<1)+xx][(y<<1)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_4x4:
	     for (y=0;y<csize>>2;y++)
	       {
		  for(x=0;x<csize>>2;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<4;yy++)
			 {
			    for(xx=0;xx<4;xx++)
			      {
				 c16->p[(x<<2)+xx][(y<<2)+yy].r=p.r;
				 c16->p[(x<<2)+xx][(y<<2)+yy].g=p.g;
				 c16->p[(x<<2)+xx][(y<<2)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_2x2:
	     for (y=0;y<csize>>3;y++)
	       {
		  for(x=0;x<csize>>3;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<8;yy++)
			 {
			    for(xx=0;xx<8;xx++)
			      {
				 c16->p[(x<<3)+xx][(y<<3)+yy].r=p.r;
				 c16->p[(x<<3)+xx][(y<<3)+yy].g=p.g;
				 c16->p[(x<<3)+xx][(y<<3)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_1x1:
	     for (y=0;y<csize>>4;y++)
	       {
		  for(x=0;x<csize>>4;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<16;yy++)
			 {
			    for(xx=0;xx<16;xx++)
			      {
				 c16->p[(x<<4)+xx][(y<<4)+yy].r=p.r;
				 c16->p[(x<<4)+xx][(y<<4)+yy].g=p.g;
				 c16->p[(x<<4)+xx][(y<<4)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   default:
	     break;
	  }
     }
   else if (csize==32)
     {
	switch(res)
	  {
	   case FLAG_RES_32x32:
	     for (y=0;y<csize;y++)
	       {
		  for(x=0;x<csize;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);	
		       c32->p[x][y].r=p.r;c32->p[x][y].g=p.g;c32->p[x][y].b=p.b;
		    }
	       }
	     break;
	   case FLAG_RES_16x16:
	     for (y=0;y<csize>>1;y++)
	       {
		  for(x=0;x<csize>>1;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<2;yy++)
			 {
			    for(xx=0;xx<2;xx++)
			      {
				 c32->p[(x<<1)+xx][(y<<1)+yy].r=p.r;
				 c32->p[(x<<1)+xx][(y<<1)+yy].g=p.g;
				 c32->p[(x<<1)+xx][(y<<1)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_8x8:
	     for (y=0;y<csize>>2;y++)
	       {
		  for(x=0;x<csize>>2;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<4;yy++)
			 {
			    for(xx=0;xx<4;xx++)
			      {
				 c32->p[(x<<2)+xx][(y<<2)+yy].r=p.r;
				 c32->p[(x<<2)+xx][(y<<2)+yy].g=p.g;
				 c32->p[(x<<2)+xx][(y<<2)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_4x4:
	     for (y=0;y<csize>>3;y++)
	       {
		  for(x=0;x<csize>>3;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<8;yy++)
			 {
			    for(xx=0;xx<8;xx++)
			      {
				 c32->p[(x<<3)+xx][(y<<3)+yy].r=p.r;
				 c32->p[(x<<3)+xx][(y<<3)+yy].g=p.g;
				 c32->p[(x<<3)+xx][(y<<3)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_2x2:
	     for (y=0;y<csize>>4;y++)
	       {
		  for(x=0;x<csize>>4;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<16;yy++)
			 {
			    for(xx=0;xx<16;xx++)
			      {
				 c32->p[(x<<4)+xx][(y<<4)+yy].r=p.r;
				 c32->p[(x<<4)+xx][(y<<4)+yy].g=p.g;
				 c32->p[(x<<4)+xx][(y<<4)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_1x1:
	     for (y=0;y<csize>>5;y++)
	       {
		  for(x=0;x<csize>>5;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<32;yy++)
			 {
			    for(xx=0;xx<32;xx++)
			      {
				 c32->p[(x<<5)+xx][(y<<5)+yy].r=p.r;
				 c32->p[(x<<5)+xx][(y<<5)+yy].g=p.g;
				 c32->p[(x<<5)+xx][(y<<5)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   default:
	     break;
	  }
     }
   else if (csize==64)
     {
	switch(res)
	  {
	   case FLAG_RES_64x64:
	     for (y=0;y<csize;y++)
	       {
		  for(x=0;x<csize;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);	
		       c64->p[x][y].r=p.r;c64->p[x][y].g=p.g;c64->p[x][y].b=p.b;
		    }
	       }
	     break;
	   case FLAG_RES_32x32:
	     for (y=0;y<csize>>1;y++)
	       {
		  for(x=0;x<csize>>1;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<2;yy++)
			 {
			    for(xx=0;xx<2;xx++)
			      {
				 c64->p[(x<<1)+xx][(y<<1)+yy].r=p.r;
				 c64->p[(x<<1)+xx][(y<<1)+yy].g=p.g;
				 c64->p[(x<<1)+xx][(y<<1)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_16x16:
	     for (y=0;y<csize>>2;y++)
	       {
		  for(x=0;x<csize>>2;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<4;yy++)
			 {
			    for(xx=0;xx<4;xx++)
			      {
				 c64->p[(x<<2)+xx][(y<<2)+yy].r=p.r;
				 c64->p[(x<<2)+xx][(y<<2)+yy].g=p.g;
				 c64->p[(x<<2)+xx][(y<<2)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_8x8:
	     for (y=0;y<csize>>3;y++)
	       {
		  for(x=0;x<csize>>3;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<8;yy++)
			 {
			    for(xx=0;xx<8;xx++)
			      {
				 c64->p[(x<<3)+xx][(y<<3)+yy].r=p.r;
				 c64->p[(x<<3)+xx][(y<<3)+yy].g=p.g;
				 c64->p[(x<<3)+xx][(y<<3)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_4x4:
	     for (y=0;y<csize>>4;y++)
	       {
		  for(x=0;x<csize>>4;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<16;yy++)
			 {
			    for(xx=0;xx<16;xx++)
			      {
				 c64->p[(x<<4)+xx][(y<<4)+yy].r=p.r;
				 c64->p[(x<<4)+xx][(y<<4)+yy].g=p.g;
				 c64->p[(x<<4)+xx][(y<<4)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_2x2:
	     for (y=0;y<csize>>5;y++)
	       {
		  for(x=0;x<csize>>5;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<32;yy++)
			 {
			    for(xx=0;xx<32;xx++)
			      {
				 c64->p[(x<<5)+xx][(y<<5)+yy].r=p.r;
				 c64->p[(x<<5)+xx][(y<<5)+yy].g=p.g;
				 c64->p[(x<<5)+xx][(y<<5)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_1x1:
	     for (y=0;y<csize>>6;y++)
	       {
		  for(x=0;x<csize>>6;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<64;yy++)
			 {
			    for(xx=0;xx<64;xx++)
			      {
				 c64->p[(x<<6)+xx][(y<<6)+yy].r=p.r;
				 c64->p[(x<<6)+xx][(y<<6)+yy].g=p.g;
				 c64->p[(x<<6)+xx][(y<<6)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   default:
	     break;
	  }
     }
   else if (csize==128)
     {
	switch(res)
	  {
	   case FLAG_RES_128x128:
	     for (y=0;y<csize;y++)
	       {
		  for(x=0;x<csize;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);	
		       c128->p[x][y].r=p.r;c128->p[x][y].g=p.g;c128->p[x][y].b=p.b;
		    }
	       }
	     break;
	   case FLAG_RES_64x64:
	     for (y=0;y<csize>>1;y++)
	       {
		  for(x=0;x<csize>>1;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<2;yy++)
			 {
			    for(xx=0;xx<2;xx++)
			      {
				 c128->p[(x<<1)+xx][(y<<1)+yy].r=p.r;
				 c128->p[(x<<1)+xx][(y<<1)+yy].g=p.g;
				 c128->p[(x<<1)+xx][(y<<1)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_32x32:
	     for (y=0;y<csize>>2;y++)
	       {
		  for(x=0;x<csize>>2;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<4;yy++)
			 {
			    for(xx=0;xx<4;xx++)
			      {
				 c128->p[(x<<2)+xx][(y<<2)+yy].r=p.r;
				 c128->p[(x<<2)+xx][(y<<2)+yy].g=p.g;
				 c128->p[(x<<2)+xx][(y<<2)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_16x16:
	     for (y=0;y<csize>>3;y++)
	       {
		  for(x=0;x<csize>>3;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<8;yy++)
			 {
			    for(xx=0;xx<8;xx++)
			      {
				 c128->p[(x<<3)+xx][(y<<3)+yy].r=p.r;
				 c128->p[(x<<3)+xx][(y<<3)+yy].g=p.g;
				 c128->p[(x<<3)+xx][(y<<3)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_8x8:
	     for (y=0;y<csize>>4;y++)
	       {
		  for(x=0;x<csize>>4;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<16;yy++)
			 {
			    for(xx=0;xx<16;xx++)
			      {
				 c128->p[(x<<4)+xx][(y<<4)+yy].r=p.r;
				 c128->p[(x<<4)+xx][(y<<4)+yy].g=p.g;
				 c128->p[(x<<4)+xx][(y<<4)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_4x4:
	     for (y=0;y<csize>>5;y++)
	       {
		  for(x=0;x<csize>>5;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<32;yy++)
			 {
			    for(xx=0;xx<32;xx++)
			      {
				 c128->p[(x<<5)+xx][(y<<5)+yy].r=p.r;
				 c128->p[(x<<5)+xx][(y<<5)+yy].g=p.g;
				 c128->p[(x<<5)+xx][(y<<5)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_2x2:
	     for (y=0;y<csize>>6;y++)
	       {
		  for(x=0;x<csize>>6;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<64;yy++)
			 {
			    for(xx=0;xx<64;xx++)
			      {
				 c128->p[(x<<6)+xx][(y<<6)+yy].r=p.r;
				 c128->p[(x<<6)+xx][(y<<6)+yy].g=p.g;
				 c128->p[(x<<6)+xx][(y<<6)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   case FLAG_RES_1x1:
	     for (y=0;y<csize>>7;y++)
	       {
		  for(x=0;x<csize>>7;x++)
		    {
		       if (code==FLAG_PIX_24b) 
			 {
			    pix=(unsigned int)byte[i++]<<16;
			    pix|=(unsigned int)byte[i++]<<8;
			    pix|=(unsigned int)byte[i++];
			 }
		       else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
				(code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
			 pix=(unsigned int)word[i++];
		       else
			 pix=(unsigned int)byte[i++];
		       decode_pixel(code,&p,pix);
		       for(yy=0;yy<128;yy++)
			 {
			    for(xx=0;xx<128;xx++)
			      {
				 c128->p[(x<<7)+xx][(y<<7)+yy].r=p.r;
				 c128->p[(x<<7)+xx][(y<<7)+yy].g=p.g;
				 c128->p[(x<<7)+xx][(y<<7)+yy].b=p.b;
			      }
			 }
		    }
	       }
	     break;
	   default:
	     break;
	  }
     }
}

void *load_chunk_chunks_m(unsigned char *m, int *size, int *w, int *h, int *x, int *y, int *csize)
{
   void *c;
   int i,code,bpp,cd,sz;
   void *data;
   char s[10];
   unsigned char *ptr;
   int *ip;
   RawChunk24_8   *c8;
   RawChunk24_16  *c16;
   RawChunk24_32  *c32;
   RawChunk24_64  *c64;
   RawChunk24_128 *c128;
   
   if (!m) 
     {
	*size=0;
	return NULL;
     }
   ptr=m+8;
   ip=(int *)ptr;
   *w=*ip;ip++;ptr+=sizeof(int);
   *h=*ip;ip++;ptr+=sizeof(int);
   *x=*ip;ip++;ptr+=sizeof(int);
   *y=*ip;ip++;ptr+=sizeof(int);
   *csize=*ip;ip++;ptr+=sizeof(int);

   if      (*csize==8)   
     {
	c8=(RawChunk24_8 *)malloc(sizeof(RawChunk24_8)**x**y);if (!c8) return NULL;
     }
   else if (*csize==16)   
     {
	c16=(RawChunk24_16 *)malloc(sizeof(RawChunk24_16)**x**y);if (!c16) return NULL;
     }
   else if (*csize==32)   
     {
	c32=(RawChunk24_32 *)malloc(sizeof(RawChunk24_32)**x**y);if (!c32) return NULL;
     }
   else if (*csize==64)   
     {
	c64=(RawChunk24_64 *)malloc(sizeof(RawChunk24_64)**x**y);if (!c64) return NULL;
     }
   else if (*csize==128)   
     {
	c128=(RawChunk24_128 *)malloc(sizeof(RawChunk24_128)**x**y);if (!c128) return NULL;
     }
   if (!c)
     {
	*size=ptr-m;
	return NULL;
     }
   for(i=0;i<*x**y;i++)
     {
	ip=(int *)ptr;
	code=*ip;ptr+=(sizeof(int));
	cd=code&FLAG_PIX_MSK;
	sz=code&FLAG_RES_MSK;
	if (cd==FLAG_PIX_24b) bpp=3;
	else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
		 (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
	else bpp=1;
	if      (sz==FLAG_RES_128x128) bpp*=128*128;
	else if (sz==FLAG_RES_64x64)   bpp*=64*64;
	else if (sz==FLAG_RES_32x32)   bpp*=32*32;
	else if (sz==FLAG_RES_16x16)   bpp*=16*16;
	else if (sz==FLAG_RES_8x8)     bpp*=8*8;
	else if (sz==FLAG_RES_4x4)     bpp*=4*4;
	else if (sz==FLAG_RES_2x2)     bpp*=2*2;
	data=(void *)ptr;
	ptr+=bpp;
	if      (*csize==8)   decode_chunk(code,data,&c8[i],*csize);
	else if (*csize==16)  decode_chunk(code,data,&c16[i],*csize);
	else if (*csize==32)  decode_chunk(code,data,&c32[i],*csize);
	else if (*csize==64)  decode_chunk(code,data,&c64[i],*csize);
	else if (*csize==128) decode_chunk(code,data,&c128[i],*csize);
     }
   *size=ptr-m;
   if      (*csize==8)   return (void *)c8;
   else if (*csize==16)  return (void *)c16;
   else if (*csize==32)  return (void *)c32;
   else if (*csize==64)  return (void *)c64;
   else if (*csize==128) return (void *)c128;
}

void *load_chunk_chunks_f(FILE *f, int *w, int *h, int *x, int *y, int *csize)
{
   int i,code,bpp,cd,sz;
   void *data;
   char s[10];
   unsigned char *v;
   int q;
   RawChunk24_8   *c8;
   RawChunk24_16  *c16;
   RawChunk24_32  *c32;
   RawChunk24_64  *c64;
   RawChunk24_128 *c128;
   
   if (!f) return NULL;
   fread(s,1,8,f);
   s[8]=0;
   fread(w,sizeof(unsigned int),1,f);
   fread(h,sizeof(unsigned int),1,f);
   fread(x,sizeof(unsigned int),1,f);
   fread(y,sizeof(unsigned int),1,f);
   fread(csize,sizeof(unsigned int),1,f);

   if      (*csize==8)   
     {
	c8=(RawChunk24_8 *)malloc(sizeof(RawChunk24_8)**x**y);if (!c8) return NULL;
     }
   else if (*csize==16)   
     {
	c16=(RawChunk24_16 *)malloc(sizeof(RawChunk24_16)**x**y);if (!c16) return NULL;
     }
   else if (*csize==32)   
     {
	c32=(RawChunk24_32 *)malloc(sizeof(RawChunk24_32)**x**y);if (!c32) return NULL;
     }
   else if (*csize==64)   
     {
	c64=(RawChunk24_64 *)malloc(sizeof(RawChunk24_64)**x**y);if (!c64) return NULL;
     }
   else if (*csize==128)   
     {
	c128=(RawChunk24_128 *)malloc(sizeof(RawChunk24_128)**x**y);if (!c128) return NULL;
     }
   data=(void *)malloc(128*128*3);
   if (!data)
     {
	if      (*csize==8)   free(c8);
	else if (*csize==16)  free(c16);
	else if (*csize==32)  free(c32);
	else if (*csize==64)  free(c64);
	else if (*csize==128) free(c128);
	return NULL;
     }
   for(i=0;i<*x**y;i++)
     {
	fread(&code,sizeof(unsigned int),1,f);
	cd=code&FLAG_PIX_MSK;sz=code&FLAG_RES_MSK;
	if (cd==FLAG_PIX_24b) bpp=3;
	else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
		 (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
	else bpp=1;
	if      (sz==FLAG_RES_128x128) bpp*=128*128;
	else if (sz==FLAG_RES_64x64)   bpp*=64*64;
	else if (sz==FLAG_RES_32x32)   bpp*=32*32;
	else if (sz==FLAG_RES_16x16)   bpp*=16*16;
	else if (sz==FLAG_RES_8x8)     bpp*=8*8;
	else if (sz==FLAG_RES_4x4)     bpp*=4*4;
	else if (sz==FLAG_RES_2x2)     bpp*=2*2;
	fread(data,bpp,1,f);
	if      (*csize==8)   decode_chunk(code,data,&c8[i],*csize);
	else if (*csize==16)  decode_chunk(code,data,&c16[i],*csize);
	else if (*csize==32)  decode_chunk(code,data,&c32[i],*csize);
	else if (*csize==64)  decode_chunk(code,data,&c64[i],*csize);
	else if (*csize==128) decode_chunk(code,data,&c128[i],*csize);
     }
   free(data);
   if      (*csize==8)   return (void *)c8;
   else if (*csize==16)  return (void *)c16;
   else if (*csize==32)  return (void *)c32;
   else if (*csize==64)  return (void *)c64;
   else if (*csize==128) return (void *)c128;
}

int load_chunks_frame_f(FILE *f, int *w, int *h, int *x, int *y, void *chk, int csize)
{
   int i,code,bpp,cd,sz;
   void *data;
   RawChunk24_8   *c8;
   RawChunk24_16  *c16;
   RawChunk24_32  *c32;
   RawChunk24_64  *c64;
   RawChunk24_128 *c128;
   
   if      (csize==8)   c8=(RawChunk24_8 *)chk;
   else if (csize==16)  c16=(RawChunk24_16 *)chk;
   else if (csize==32)  c32=(RawChunk24_32 *)chk;
   else if (csize==64)  c64=(RawChunk24_64 *)chk;
   else if (csize==128) c128=(RawChunk24_128 *)chk;
   data=(void *)malloc(128*128*3);
   if (!data) return 0;
   if (!fread(&code,sizeof(unsigned int),1,f)) return -1;
   if (code==0xffffffff) 
     {
	free(data);
	return 0;
     }
   if (code==0xfffffffe) 
     {
	free(data);
	return -2;
     }
   cd=code&FLAG_PIX_MSK;sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if      (sz==FLAG_RES_128x128) bpp*=128*128;
   else if (sz==FLAG_RES_64x64)   bpp*=64*64;
   else if (sz==FLAG_RES_32x32)   bpp*=32*32;
   else if (sz==FLAG_RES_16x16)   bpp*=16*16;
   else if (sz==FLAG_RES_8x8)     bpp*=8*8;
   else if (sz==FLAG_RES_4x4)     bpp*=4*4;
   else if (sz==FLAG_RES_2x2)     bpp*=2*2;
   if (!fread(data,bpp,1,f)) 
     {
	free(data);
	fprintf(stderr,"eof\n");
	return -1;
     }
   i=code>>7;
   if      (csize==8)   decode_chunk(code,data,&c8[i],csize);
   else if (csize==16)  decode_chunk(code,data,&c16[i],csize);
   else if (csize==32)  decode_chunk(code,data,&c32[i],csize);
   else if (csize==64)  decode_chunk(code,data,&c64[i],csize);
   else if (csize==128) decode_chunk(code,data,&c128[i],csize);
   free(data);
   return i+1;
}

unsigned char *memopen(char *file, int *size)
{
   struct stat st;
   void *buf;
   int fd;

   stat(file,&st);
   buf=(void *)malloc(st.st_size);
   if (!buf) return NULL;
   fd=open(file,O_RDONLY);
   read(fd,buf,st.st_size);
   close(fd);
   return buf;
}

int main(int argc, char **argv)
{
   void *c,*c1,*c2;
   int w,h,x,y;
   int i,code,size,ii,j,sz;
   void *data;
   void *buf;
   unsigned int maxloss,maxdelta,loss;
   unsigned char *ppmdata;
   FILE *f,*ff;
   unsigned char *m,*mm;
   char s[1024];
   char *block;
   int loop;
   int mem,pos;
   int glob,root;
   int cnt;
   struct timeval tv;
   double t1,t2,tt,tot;
   struct sigaction sa;
   char file[1024];
   int csize;
   
   shmp=0;
   glob=2;
   mem=1;
   loop=1;
   root=0;
   sa.sa_handler = qit;
   sa.sa_flags = 0;
   sigemptyset (&sa.sa_mask);
   sigaction(SIGHUP,&sa,(struct sigaction *)0);
   sigaction(SIGINT,&sa,(struct sigaction *)0);
   sigaction(SIGQUIT,&sa,(struct sigaction *)0);
   sigaction(SIGTERM,&sa,(struct sigaction *)0);
   sigaction(SIGABRT,&sa,(struct sigaction *)0);
   sigaction(SIGPIPE,&sa,(struct sigaction *)0);
   
   if (argc<2)
     {
	printf("usage:\n\t%s anim_file [-file] [-noloop] [-glob n] [-root] [-shmp]\n",argv[0]);
	printf("\t-file : Play file off disk rather than memory\n");
	printf("\t-noloop : Do not loop animation (play once then exit)\n");
	printf("\t-glob n: Set display update globsize to n\n");
	printf("\t-root : Play animation in root window\n");
	printf("\t-shmp : Use shared pixmaps instead of shared images\n");
	exit(0);
     }
   if (argc>1)
     {
	if (!strcmp("-h",argv[1]))
	  {
	     printf("usage:\n\t%s anim_file [-file] [-noloop] [-glob n] [-root] [-shmp]\n",argv[0]);
	     printf("\t-file : Play file off disk rather than memory\n");
	     printf("\t-noloop : Do not loop animation (play once then exit)\n");
	     printf("\t-glob n: Set display update globsize to n\n");
	     printf("\t-root : Play animation in root window\n");
	     printf("\t-shmp : Use shared pixmaps instead of shared images\n");
	     exit(0);
	  }
     }
   if (argc>1)
     {
	strcpy(file,argv[1]);
	for(i=2;i<argc;i++)
	  {
	     if (!strcmp("-file",argv[i])) mem=0;
	     else if (!strcmp("-noloop",argv[i])) loop=0;
	     else if (!strcmp("-root",argv[i])) root=1;
	     else if (!strcmp("-shmp",argv[i])) shmp=1;
	     else if (!strcmp("-glob",argv[i++])) glob=atoi(argv[i]);
	     else if (!strcmp("-h",argv[i]))
	       {
		  printf("usage:\n\t%s anim_file [-file] [-noloop] [-glob n]\n");
		  printf("\t-file : Play file off disk rather than memory\n");
		  printf("\t-noloop : Do not loop animation (play once then exit)\n");
		  printf("\t-glob n: Set display update globsize to n\n");
		  printf("\t-root n: Play animation in root window\n");
		  exit(0);
	       }
	  }
     }
   if (argc>=0)
     {
	if (mem) 
	  {
	     m=memopen(file,&size);
	     mm=m;
	     c=load_chunk_chunks_m(mm,&sz,&w,&h,&x,&y,&csize);
	     mm+=sz;
	  }
	else
	  {
	     f=fopen(file,"r");
	     c=load_chunk_chunks_f(f,&w,&h,&x,&y,&csize);
	  }
	block=(char *)malloc(x*y);
	X_OpenWindow(0,0,w,h,root);
	X_SetupImage(w,h);
	X_Sync();
	i=1;
	X_Dechunkify(w,h,x,y,c,csize);
	free(c);
	X_Sync();
	i++;
	j=0;
	cnt=0;
	for (i=0;i<x*y;i++) block[i]=0;
	if (mem)
	  {
	     tot=0;
	     while ((ii=X_Load_chunks_frame_m(mm,&sz,&w,&h,&x,&y,csize))>-3)
	       {
		  if (ii>0) block[ii-1]=1;
		  mm+=sz;
		  if (ii<=0)
		    {
		       j++;
		       X_Sync();
		       cnt+=X_DisplayImage(block,x,y,glob,csize);
		       for (i=0;i<x*y;i++) block[i]=0;
		       gettimeofday(&tv,NULL);
		       t2=(double)tv.tv_sec+((double)tv.tv_usec/1000000);
		       tt=t2-t1;
		       tot+=tt;
		       t1=t2;
		       if (ii<0)
			 {
			    printf("frames %i curfps: %2.2f fps: %2.2f deltaavg: %3.2f%%\r",j,1/tt,(double)j/tot,(double)cnt*100/(double)(j*x*y));
			    fflush(stdout);
			 }
		    }
		  if (ii==-1) 
		    {
		       break;
		    }
		  else if (ii==-2)
		    {
		       mm=m+28;
		       X_Sync();
		       tot=0;
		       gettimeofday(&tv,NULL);
		       t1=(double)tv.tv_sec+((double)tv.tv_usec/1000000);
		       j=1;
		       cnt=0;
		       while (X_Skip_chunks_frame_m(mm,&sz)>0) mm+=sz;
		       mm+=sz;
		    }
	       }
	  }
	else
	  {
	     while ((ii=X_Load_chunks_frame_f(f,&w,&h,&x,&y,csize))>-3)
	       {
		  if (ii>0) block[ii-1]=1;
		  if (ii<=0)
		    {
		       printf("%i\n",j);
		       j++;
		       X_Sync();
		       cnt+=X_DisplayImage(block,x,y,glob,csize);
		       for (i=0;i<x*y;i++) block[i]=0;
		       gettimeofday(&tv,NULL);
		       t2=(double)tv.tv_sec+((double)tv.tv_usec/1000000);
		       tt=t2-t1;
		       tot+=tt;
		       t1=t2;
		       if (ii<0)
			 {
			    printf("frames %i curfps: %2.2f fps: %2.2f deltaavg: %3.2f%%\r",j,1/tt,(double)j/tot,(double)cnt*100/(double)(j*x*y));
			    fflush(stdout);
			 }
		    }
		  if ((ii==-1)&&(!loop)) 
		    {
		       break;
		    }
		  else if ((ii==-2)&&(loop))
		    {
		       fseek(f,SEEK_SET,0);
		       fread(s,1,8,f);
		       fread(&w,sizeof(unsigned int),1,f);
		       fread(&h,sizeof(unsigned int),1,f);
		       fread(&x,sizeof(unsigned int),1,f);
		       fread(&y,sizeof(unsigned int),1,f);
		       fread(&csize,sizeof(unsigned int),1,f);
		       X_Sync();
		       tot=0;
		       gettimeofday(&tv,NULL);
		       t1=(double)tv.tv_sec+((double)tv.tv_usec/1000000);
/*		       while (X_Skip_chunks_frame_f(f,&sz)>0);*/
		       j=0;
		       cnt=0;
		    }
	       }
	     fclose(f);
	  }
     }
}
