#include "eanim.h"

void decode_pixel(int code, Pixel24 *p, int pix);

Display *disp;
Window win;
Visual *vis;
int scr;
GC gc;
int depth;
int winwidth;
int winheight;
int imgjmp;
char shmp;
Colormap cmap;
XImage *xim;
XShmSegmentInfo shminfo;
Pixmap pmap;


void qit()
{
   printf("\n\n\nClean Exit\n\n\n");
   XShmDetach(disp,&shminfo);
   shmdt(shminfo.shmaddr);
   shmctl(shminfo.shmid,IPC_RMID,0);
   XDestroyImage(xim);
   XFreePixmap(disp,pmap);
   XSetWindowBackgroundPixmap(disp,win,0);
   XSetWindowBackground(disp,win,0);
   XClearWindow(disp,win);
   exit(0);
}

void X_SetupImage(int w,int h)
{
   xim=XShmCreateImage(disp,vis,depth,ZPixmap,NULL,&shminfo,w,h);
   if (!xim)
     fprintf(stderr,"X-Error: unable to create XShm XImage\n");
   shminfo.shmid=shmget(IPC_PRIVATE,
			xim->bytes_per_line*xim->height,
			IPC_CREAT|0777);
   if (shminfo.shmid==-1)
     fprintf(stderr,"SharedMemory Error: unable to get identifier\n");
   shminfo.shmaddr=xim->data=shmat(shminfo.shmid,0,0);
   if(!XShmAttach(disp,&shminfo))
     fprintf(stderr,"X-Error: unable to attach XShm Shared Memory Segment\n");
   if (shmp)
     pmap=XShmCreatePixmap(disp,win,shminfo.shmaddr,&shminfo,w,h,depth);
   else
     pmap=XCreatePixmap(disp,win,w,h,depth);
   if (!pmap)
     fprintf(stderr,"Unable to create pixmap\n");
   imgjmp=w;
   XSetWindowBackgroundPixmap(disp,win,pmap);
}

int X_DisplayImage(char *block, int bx, int by, int glob)
{
   int x,y,xx,yy,gx,gy;
   int show,i,cnt;
   
   cnt=0;
   if (block)
     {
	for (y=0;y<by;y+=glob)
	  {
	     for (x=0;x<bx;x+=glob)
	       {
		  show=0;
		  if ((bx-x)<glob) gx=bx-x;
		  else gx=glob;
		  if ((by-y)<glob) gy=by-y;
		  else gy=glob;
		  for (yy=0;yy<gy;yy++)
		    {
		       i=((y+yy)*bx)+x;
		       for (xx=0;xx<gx;xx++) show+=block[i++];
		    }
		  if (show)
		    {
		       if (!shmp)
			 XShmPutImage(disp,pmap,gc,xim,x<<3,y<<3,x<<3,y<<3,
				      gx<<3,gy<<3,False);
		       XClearArea(disp,win,x<<3,y<<3,gx<<3,gy<<3,False);
		       cnt+=glob*glob;
		    }
	       }
	  }
     }
   else
     {
	if (!shmp)
	  XShmPutImage(disp,pmap,gc,xim,0,0,0,0,winwidth,winheight,False);
	XClearWindow(disp,win);
     }
   return cnt;
}

void X_OpenWindow(int x, int y, int w,int h, int rr)
{
   XGCValues gcvals;
   Window root;
   XColor xcl;
   int i;
   int r,g,b;
   
   disp=XOpenDisplay(NULL);
   if (disp)
     {
	scr=DefaultScreen(disp);
	vis=DefaultVisual(disp,scr);
	root=DefaultRootWindow(disp);
	depth=DefaultDepth(disp,scr);
	if (!rr) win=XCreateSimpleWindow(disp,root,x,y,w,h,0,0,0);
	else win=root;
	gc=XCreateGC(disp,win,(unsigned long)0,&gcvals);
	XSetForeground(disp,gc,0);
	XSetBackground(disp,gc,1);
	if (!rr) XMapWindow(disp,win);
	XFlush(disp);
	winwidth=w;
	winheight=h;
     }
}

void X_Sync()
{
   XSync(disp,False);
}

void X_Dechunkify(int w, int h, int x, int y, RawChunk24 *c)
{
   int xx,yy,x2,y2,i,j,val;
   
   i=0;
   for(yy=0;yy<y;yy++)
     {
	for(xx=0;xx<x;xx++)
	  {
	     for(y2=0;y2<8;y2++)
	       {
		  for(x2=0;x2<8;x2++)
		    {
		       val=(c[i].p[x2][y2].r<<8)&0xf800|
			 (c[i].p[x2][y2].g<<3)&0x07e0|
			 (c[i].p[x2][y2].b>>3)&0x001f;
		       XPutPixel(xim,(xx<<3)+x2,(yy<<3)+y2,val);
		    }
	       }
	     i++;
	  }
     }
   X_DisplayImage(NULL,0,0,0);
}

void X_Decode_chunk3(int code_type, void *data, int x2, int y2, int w, int h)
{
   Pixel24 p;
   int bpp,x,y,xx,yy,res,i,code,val;
   int x1,y1,xx1,yy1;
   unsigned int pix;
   unsigned char *byte;
   unsigned short *word;
   unsigned short *ptr;
   int jmp,jmp2,jmp4,jmpb1,jmp_2,jmpb3,jmp_4;
   
   res=code_type&FLAG_RES_MSK;
   code=code_type&FLAG_PIX_MSK;
   if (code==FLAG_PIX_24b) bpp=3;
   else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
	    (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (bpp==2) word=(unsigned short *)data;
   else byte=(unsigned char *)data;
   i=0;
   ptr=(unsigned short *)xim->data;
   ptr+=((w<<3)*y2)+(x2<<3);
/* defines for evaluating pixel valuse for 16bpp 565 */
#define VAL_24b p.r=byte[i++];p.g=byte[i++];p.b=byte[i++];val=((p.r<<8)&0xf800)|((p.g<<3)&0x07e0)|((p.b>>3)&0x001f)
#define VAL_16b val=(word[i++])
#define VAL_RG  val=((word[i])&0xf800)|((word[i++]<<3)&0x07e0)
#define VAL_RB  val=((word[i])&0xf800)|((word[i++]>>3)&0x001f)
#define VAL_GB  val=((word[i]>>5)&0x07e0)|((word[i++]>>3)&0x001f)
#define VAL_332 val=((byte[i]<<8)&0xe000)|((byte[i]<<6)&0x0700)|((byte[i++]<<3)&0x0018)
#define VAL_R   val=((byte[i++]<<8)&0xf800)
#define VAL_G   val=((byte[i++]<<3)&0x07e0)
#define VAL_B   val=((byte[i++]>>3)&0x001f)
#define VAL_RG8 val=((byte[i]<<8)&0xf000)|((byte[i++]<<7)&0x0780)
#define VAL_RB8 val=((byte[i]<<8)&0xf000)|((byte[i++]<<1)&0x001e)
#define VAL_GB8 val=((byte[i]<<3)&0x0780)|((byte[i++]<<1)&0x001e)
#define VAL_GRY val=((byte[i]<<8)&0xf800)|((byte[i]<<3)&0x07e0)|((byte[i++]>>3)&0x001f)
#define VAL_422 val=((byte[i]<<8)&0xf000)|((byte[i]<<7)&0x0600)|((byte[i++]<<3)&0x0018)
#define VAL_242 val=((byte[i]<<8)&0xc000)|((byte[i]<<5)&0x0780)|((byte[i++]<<3)&0x0018)
#define VAL_224 val=((byte[i]<<8)&0xc000)|((byte[i]<<5)&0x0600)|((byte[i++]<<1)&0x001e)
   
#define SET_11PIX *ptr++=val
#define SET_22PIX *ptr++=val;*ptr=val;ptr+=jmp2;\
                  *ptr++=val;*ptr=val;ptr-=jmpb1
#define SET_44PIX *ptr++=val;*ptr++=val;*ptr++=val;*ptr=val;ptr+=jmp4;\
                  *ptr++=val;*ptr++=val;*ptr++=val;*ptr=val;ptr+=jmp4;\
                  *ptr++=val;*ptr++=val;*ptr++=val;*ptr=val;ptr+=jmp4;\
                  *ptr++=val;*ptr++=val;*ptr++=val;*ptr=val;ptr-=jmpb3
#define LIN_1JMP  ptr+=jmp
#define LIN_2JMP  ptr+=jmp_2
#define LIN_4JMP  ptr+=jmp_4   

   switch(res)
     {
      case FLAG_RES_8x8:
	y1=y2<<3;yy1=y1+8;
	jmp=w-8;
	switch (code)
	  {
	   case FLAG_PIX_24b:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_24b;SET_11PIX;
		  VAL_24b;SET_11PIX;
		  VAL_24b;SET_11PIX;
		  VAL_24b;SET_11PIX;
		  VAL_24b;SET_11PIX;
		  VAL_24b;SET_11PIX;
		  VAL_24b;SET_11PIX;
		  VAL_24b;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_16b:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_16b;SET_11PIX;
		  VAL_16b;SET_11PIX;
		  VAL_16b;SET_11PIX;
		  VAL_16b;SET_11PIX;
		  VAL_16b;SET_11PIX;
		  VAL_16b;SET_11PIX;
		  VAL_16b;SET_11PIX;
		  VAL_16b;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_RG:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_RG;SET_11PIX;
		  VAL_RG;SET_11PIX;
		  VAL_RG;SET_11PIX;
		  VAL_RG;SET_11PIX;
		  VAL_RG;SET_11PIX;
		  VAL_RG;SET_11PIX;
		  VAL_RG;SET_11PIX;
		  VAL_RG;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_RB:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_RB;SET_11PIX;
		  VAL_RB;SET_11PIX;
		  VAL_RB;SET_11PIX;
		  VAL_RB;SET_11PIX;
		  VAL_RB;SET_11PIX;
		  VAL_RB;SET_11PIX;
		  VAL_RB;SET_11PIX;
		  VAL_RB;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_GB:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_GB;SET_11PIX;
		  VAL_GB;SET_11PIX;
		  VAL_GB;SET_11PIX;
		  VAL_GB;SET_11PIX;
		  VAL_GB;SET_11PIX;
		  VAL_GB;SET_11PIX;
		  VAL_GB;SET_11PIX;
		  VAL_GB;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_332:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_332;SET_11PIX;
		  VAL_332;SET_11PIX;
		  VAL_332;SET_11PIX;
		  VAL_332;SET_11PIX;
		  VAL_332;SET_11PIX;
		  VAL_332;SET_11PIX;
		  VAL_332;SET_11PIX;
		  VAL_332;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_R:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_R;SET_11PIX;
		  VAL_R;SET_11PIX;
		  VAL_R;SET_11PIX;
		  VAL_R;SET_11PIX;
		  VAL_R;SET_11PIX;
		  VAL_R;SET_11PIX;
		  VAL_R;SET_11PIX;
		  VAL_R;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_G:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_G;SET_11PIX;
		  VAL_G;SET_11PIX;
		  VAL_G;SET_11PIX;
		  VAL_G;SET_11PIX;
		  VAL_G;SET_11PIX;
		  VAL_G;SET_11PIX;
		  VAL_G;SET_11PIX;
		  VAL_G;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_B:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_B;SET_11PIX;
		  VAL_B;SET_11PIX;
		  VAL_B;SET_11PIX;
		  VAL_B;SET_11PIX;
		  VAL_B;SET_11PIX;
		  VAL_B;SET_11PIX;
		  VAL_B;SET_11PIX;
		  VAL_B;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_RG8:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_RG8;SET_11PIX;
		  VAL_RG8;SET_11PIX;
		  VAL_RG8;SET_11PIX;
		  VAL_RG8;SET_11PIX;
		  VAL_RG8;SET_11PIX;
		  VAL_RG8;SET_11PIX;
		  VAL_RG8;SET_11PIX;
		  VAL_RG8;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_RB8:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_RB8;SET_11PIX;
		  VAL_RB8;SET_11PIX;
		  VAL_RB8;SET_11PIX;
		  VAL_RB8;SET_11PIX;
		  VAL_RB8;SET_11PIX;
		  VAL_RB8;SET_11PIX;
		  VAL_RB8;SET_11PIX;
		  VAL_RB8;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_GB8:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_GB8;SET_11PIX;
		  VAL_GB8;SET_11PIX;
		  VAL_GB8;SET_11PIX;
		  VAL_GB8;SET_11PIX;
		  VAL_GB8;SET_11PIX;
		  VAL_GB8;SET_11PIX;
		  VAL_GB8;SET_11PIX;
		  VAL_GB8;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_GRY:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_GRY;SET_11PIX;
		  VAL_GRY;SET_11PIX;
		  VAL_GRY;SET_11PIX;
		  VAL_GRY;SET_11PIX;
		  VAL_GRY;SET_11PIX;
		  VAL_GRY;SET_11PIX;
		  VAL_GRY;SET_11PIX;
		  VAL_GRY;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_422:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_422;SET_11PIX;
		  VAL_422;SET_11PIX;
		  VAL_422;SET_11PIX;
		  VAL_422;SET_11PIX;
		  VAL_422;SET_11PIX;
		  VAL_422;SET_11PIX;
		  VAL_422;SET_11PIX;
		  VAL_422;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_242:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_242;SET_11PIX;
		  VAL_242;SET_11PIX;
		  VAL_242;SET_11PIX;
		  VAL_242;SET_11PIX;
		  VAL_242;SET_11PIX;
		  VAL_242;SET_11PIX;
		  VAL_242;SET_11PIX;
		  VAL_242;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   case FLAG_PIX_224:
	     for (y=y1;y<yy1;y++)
	       {
		  VAL_224;SET_11PIX;
		  VAL_224;SET_11PIX;
		  VAL_224;SET_11PIX;
		  VAL_224;SET_11PIX;
		  VAL_224;SET_11PIX;
		  VAL_224;SET_11PIX;
		  VAL_224;SET_11PIX;
		  VAL_224;SET_11PIX;
		  LIN_1JMP;
	       }
	     break;
	   default:
	     break;
	  }
	break;
      case FLAG_RES_4x4:
	y1=y2<<3;yy1=y1+8;
	jmp2=(w-2)+1;
	jmpb1=w-1;
	jmp_2=(w<<1)-8;
	switch (code)
	  {
	   case FLAG_PIX_24b:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_24b;SET_22PIX;
		  VAL_24b;SET_22PIX;
		  VAL_24b;SET_22PIX;
		  VAL_24b;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_16b:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_16b;SET_22PIX;
		  VAL_16b;SET_22PIX;
		  VAL_16b;SET_22PIX;
		  VAL_16b;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_RG:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_RG;SET_22PIX;
		  VAL_RG;SET_22PIX;
		  VAL_RG;SET_22PIX;
		  VAL_RG;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_RB:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_RB;SET_22PIX;
		  VAL_RB;SET_22PIX;
		  VAL_RB;SET_22PIX;
		  VAL_RB;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_GB:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_GB;SET_22PIX;
		  VAL_GB;SET_22PIX;
		  VAL_GB;SET_22PIX;
		  VAL_GB;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_332:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_332;SET_22PIX;
		  VAL_332;SET_22PIX;
		  VAL_332;SET_22PIX;
		  VAL_332;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_R:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_R;SET_22PIX;
		  VAL_R;SET_22PIX;
		  VAL_R;SET_22PIX;
		  VAL_R;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_G:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_G;SET_22PIX;
		  VAL_G;SET_22PIX;
		  VAL_G;SET_22PIX;
		  VAL_G;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_B:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_B;SET_22PIX;
		  VAL_B;SET_22PIX;
		  VAL_B;SET_22PIX;
		  VAL_B;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_RG8:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_RG8;SET_22PIX;
		  VAL_RG8;SET_22PIX;
		  VAL_RG8;SET_22PIX;
		  VAL_RG8;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_RB8:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_RB8;SET_22PIX;
		  VAL_RB8;SET_22PIX;
		  VAL_RB8;SET_22PIX;
		  VAL_RB8;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_GB8:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_GB8;SET_22PIX;
		  VAL_GB8;SET_22PIX;
		  VAL_GB8;SET_22PIX;
		  VAL_GB8;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_GRY:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_GRY;SET_22PIX;
		  VAL_GRY;SET_22PIX;
		  VAL_GRY;SET_22PIX;
		  VAL_GRY;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_422:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_422;SET_22PIX;
		  VAL_422;SET_22PIX;
		  VAL_422;SET_22PIX;
		  VAL_422;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_242:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_242;SET_22PIX;
		  VAL_242;SET_22PIX;
		  VAL_242;SET_22PIX;
		  VAL_242;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   case FLAG_PIX_224:
	     for (y=y1;y<yy1;y+=2)
	       {
		  VAL_224;SET_22PIX;
		  VAL_224;SET_22PIX;
		  VAL_224;SET_22PIX;
		  VAL_224;SET_22PIX;
		  LIN_2JMP;
	       }
	     break;
	   default:
	     break;
	  }
	break;
      case FLAG_RES_2x2:
	jmp4=(w-4)+1;
	jmpb3=((w<<1)+w)-1;
	jmp_4=(w<<2)-8;
	switch (code)
	  {
	   case FLAG_PIX_24b:
	     VAL_24b;SET_44PIX;
	     VAL_24b;SET_44PIX;
	     LIN_4JMP;
	     VAL_24b;SET_44PIX;
	     VAL_24b;SET_44PIX;
	     break;
	   case FLAG_PIX_16b:
	     VAL_16b;SET_44PIX;
	     VAL_16b;SET_44PIX;
	     LIN_4JMP;
	     VAL_16b;SET_44PIX;
	     VAL_16b;SET_44PIX;
	     break;
	   case FLAG_PIX_RG:
	     VAL_RG;SET_44PIX;
	     VAL_RG;SET_44PIX;
	     LIN_4JMP;
	     VAL_RG;SET_44PIX;
	     VAL_RG;SET_44PIX;
	     break;
	   case FLAG_PIX_RB:
	     VAL_RB;SET_44PIX;
	     VAL_RB;SET_44PIX;
	     LIN_4JMP;
	     VAL_RB;SET_44PIX;
	     VAL_RB;SET_44PIX;
	     break;
	   case FLAG_PIX_GB:
	     VAL_GB;SET_44PIX;
	     VAL_GB;SET_44PIX;
	     LIN_4JMP;
	     VAL_GB;SET_44PIX;
	     VAL_GB;SET_44PIX;
	     break;
	   case FLAG_PIX_332:
	     VAL_332;SET_44PIX;
	     VAL_332;SET_44PIX;
	     LIN_4JMP;
	     VAL_332;SET_44PIX;
	     VAL_332;SET_44PIX;
	     break;
	   case FLAG_PIX_R:
	     VAL_R;SET_44PIX;
	     VAL_R;SET_44PIX;
	     LIN_4JMP;
	     VAL_R;SET_44PIX;
	     VAL_R;SET_44PIX;
	     break;
	   case FLAG_PIX_G:
	     VAL_G;SET_44PIX;
	     VAL_G;SET_44PIX;
	     LIN_4JMP;
	     VAL_G;SET_44PIX;
	     VAL_G;SET_44PIX;
	     break;
	   case FLAG_PIX_B:
	     VAL_B;SET_44PIX;
	     VAL_B;SET_44PIX;
	     LIN_4JMP;
	     VAL_B;SET_44PIX;
	     VAL_B;SET_44PIX;
	     break;
	   case FLAG_PIX_RG8:
	     VAL_RG8;SET_44PIX;
	     VAL_RG8;SET_44PIX;
	     LIN_4JMP;
	     VAL_RG8;SET_44PIX;
	     VAL_RG8;SET_44PIX;
	     break;
	   case FLAG_PIX_RB8:
	     VAL_RB8;SET_44PIX;
	     VAL_RB8;SET_44PIX;
	     LIN_4JMP;
	     VAL_RB8;SET_44PIX;
	     VAL_RB8;SET_44PIX;
	     break;
	   case FLAG_PIX_GB8:
	     VAL_GB8;SET_44PIX;
	     VAL_GB8;SET_44PIX;
	     LIN_4JMP;
	     VAL_GB8;SET_44PIX;
	     VAL_GB8;SET_44PIX;
	     break;
	   case FLAG_PIX_GRY:
	     VAL_GRY;SET_44PIX;
	     VAL_GRY;SET_44PIX;
	     LIN_4JMP;
	     VAL_GRY;SET_44PIX;
	     VAL_GRY;SET_44PIX;
	     break;
	   case FLAG_PIX_422:
	     VAL_422;SET_44PIX;
	     VAL_422;SET_44PIX;
	     LIN_4JMP;
	     VAL_422;SET_44PIX;
	     VAL_422;SET_44PIX;
	     break;
	   case FLAG_PIX_242:
	     VAL_242;SET_44PIX;
	     VAL_242;SET_44PIX;
	     LIN_4JMP;
	     VAL_242;SET_44PIX;
	     VAL_242;SET_44PIX;
	     break;
	   case FLAG_PIX_224:
	     VAL_224;SET_44PIX;
	     VAL_224;SET_44PIX;
	     LIN_4JMP;
	     VAL_224;SET_44PIX;
	     VAL_224;SET_44PIX;
	     break;
	   default:
	     break;
	  }
	break;
      case FLAG_RES_1x1:
	jmp4=(w-4)+1;
	jmpb3=((w<<1)+w)-1;
	jmp_4=(w<<2)-8;
	switch (code)
	  {
	   case FLAG_PIX_24b:
	     VAL_24b;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_16b:
	     VAL_16b;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_RG:
	     VAL_RG;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_RB:
	     VAL_RB;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_GB:
	     VAL_GB;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_332:
	     VAL_332;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_R:
	     VAL_R;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_G:
	     VAL_G;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_B:
	     VAL_B;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_RG8:
	     VAL_RG8;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_RB8:
	     VAL_RB8;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_GB8:
	     VAL_GB8;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_GRY:
	     VAL_GRY;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_422:
	     VAL_422;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_242:
	     VAL_242;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   case FLAG_PIX_224:
	     VAL_224;SET_44PIX;SET_44PIX;
	     LIN_4JMP;
	     SET_44PIX;SET_44PIX;
	     break;
	   default:
	     break;
	  }
	break;
      default:
	break;
     }
}

int X_Load_chunks_frame_m(unsigned char *m, int *size, int *w, int *h, int *x, int *y, RawChunk24 *c)
{
   int i,code,bpp,cd,sz,xx,yy;
   void *data;
   int *ip;
   unsigned char *ptr;
   
   ptr=m;
   ip=(int *)ptr;
   code=*ip;ptr+=sizeof(int);
   if (code==0xffffffff) 
     {
	*size=ptr-m;
	return 0;
     }
   else if (code==0xfffffffe) 
     {
	*size=ptr-m;
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (sz==FLAG_RES_8x8) bpp*=64;
   else if (sz==FLAG_RES_4x4) bpp*=16;
   else if (sz==FLAG_RES_2x2) bpp*=4;
   data=(void *)ptr;
   ptr+=bpp;
   i=(code>>6)&0x03ffffff;
   xx=i%*x;
   yy=(i/(*x));
   X_Decode_chunk3(code,data,xx,yy,*w,*h);
   *size=ptr-m;
   return i+1;
}

int X_Skip_chunks_frame_m(unsigned char *m, int *size, int *w, int *h, int *x, int *y, RawChunk24 *c)
{
   int i,code,bpp,cd,sz,xx,yy;
   void *data;
   int *ip;
   unsigned char *ptr;
   
   ptr=m;
   ip=(int *)ptr;
   code=*ip;ptr+=sizeof(int);
   if (code==0xffffffff) 
     {
	*size=ptr-m;
	return 0;
     }
   else if (code==0xfffffffe) 
     {
	*size=ptr-m;
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (sz==FLAG_RES_8x8) bpp*=64;
   else if (sz==FLAG_RES_4x4) bpp*=16;
   else if (sz==FLAG_RES_2x2) bpp*=4;
   ptr+=bpp;
   *size=ptr-m;
   return i+1;
}

int X_Skip_chunks_frame_f(FILE *f, int *size, int *w, int *h, int *x, int *y, RawChunk24 *c)
{
   int i,code,bpp,cd,sz,xx,yy;
   void *data;
   int *ip;
   unsigned char *ptr;
   
   if (!fread(&code,sizeof(unsigned int),1,f))
     {
	return -1;
     }
   if (code==0xffffffff) 
     {
	*size=sizeof(unsigned int);
	return 0;
     }
   else if (code==0xfffffffe) 
     {
	*size=sizeof(unsigned int);
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (sz==FLAG_RES_8x8) bpp*=64;
   else if (sz==FLAG_RES_4x4) bpp*=16;
   else if (sz==FLAG_RES_2x2) bpp*=4;
   fseek(f,bpp,SEEK_CUR);
   *size=sizeof(unsigned int)+bpp;
   return i+1;
}

int X_Load_chunks_frame_f(FILE *f, int *w, int *h, int *x, int *y, RawChunk24 *c)
{
   int i,code,bpp,cd,sz,xx,yy;
   void *data;

   data=(void *)malloc(64*3);
   if (!data)
     {
	return 0;
     }
   if (!fread(&code,sizeof(unsigned int),1,f))
     {
	return -1;
     }
   if (code==0xffffffff) 
     {
	free(data);
	return 0;
     }
   else if (code==0xfffffffe) 
     {
	free(data);
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (sz==FLAG_RES_8x8) bpp*=64;
   else if (sz==FLAG_RES_4x4) bpp*=16;
   else if (sz==FLAG_RES_2x2) bpp*=4;
   if (!fread(data,bpp,1,f)) 
     {
	fprintf(stderr,"eof\n");
	return -1;
     }
   i=(code>>6)&0x03ffffff;
   xx=i%*x;
   yy=(i/(*x));
   X_Decode_chunk3(code,data,xx,yy,*w,*h);
   free(data);
   return i+1;
}

void decode_pixel(int code, Pixel24 *p, int pix)
{
   unsigned int r,g,b;
   
   switch (code)
     {
      case FLAG_PIX_24b:
	r=(pix>>16)&0xff;
	g=(pix>>8)&0xff;
	b=(pix)&0xff;
	break;
      case FLAG_PIX_16b:
	r=(pix>>8)&0xf8;
	g=(pix>>3)&0xfc;
	b=(pix<<3)&0xf8;
	break;
      case FLAG_PIX_332:
	r=(pix)&0xe0;
	g=(pix<<3)&0xe0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_R:
	r=pix;
	g=0;
	b=0;
	break;
      case FLAG_PIX_G:
	r=0;
	g=pix;
	b=0;
	break;
      case FLAG_PIX_B:
	r=0;
	g=0;
	b=pix;
	break;
      case FLAG_PIX_RG:
	r=(pix>>8);
	g=pix&0xff;
	b=0;
	break;
      case FLAG_PIX_RB:
	r=(pix>>8);
	g=0;
	b=pix&0xff;
	break;
      case FLAG_PIX_GB:
	r=0;
	g=(pix>>8);
	b=pix&0xff;
	break;
      case FLAG_PIX_RG8:
	r=pix&0xf0;
	g=(pix<<4)&0xf0;
	b=0;
	break;
      case FLAG_PIX_RB8:
	r=pix&0xf0;
	g=0;
	b=(pix<<4)&0xf0;
	break;
      case FLAG_PIX_GB8:
	r=0;
	g=pix&0xf0;
	b=(pix<<4)&0xf0;
	break;
      case FLAG_PIX_GRY:
	r=pix;
	g=pix;
	b=pix;
	break;
      case FLAG_PIX_422:
	r=pix&0xf0;
	g=(pix<<4)&0xc0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_242:
	r=pix&0xc0;
	g=(pix<<2)&0xf0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_224:
	r=pix&0xc0;
	g=(pix<<2)&0xc0;
	b=(pix<<4)&0xf0;
	break;
      default:
	break;
     }
   p->r=(unsigned char)r;
   p->g=(unsigned char)g;
   p->b=(unsigned char)b;
}

void decode_chunk(int code_type, void *data, RawChunk24 *c)
{
   Pixel24 p;
   int bpp,x,y,xx,yy,res,i,code;
   unsigned int pix;
   unsigned char *byte;
   unsigned short *word;
   
   res=code_type&FLAG_RES_MSK;
   code=code_type&FLAG_PIX_MSK;
   if (code==FLAG_PIX_24b) bpp=3;
   else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
	    (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (bpp==2) word=(unsigned short *)data;
   else byte=(unsigned char *)data;
   i=0;
   switch(res)
     {
      case FLAG_RES_8x8:
	for (y=0;y<8;y++)
	  {
	     for(x=0;x<8;x++)
	       {
		  if (code==FLAG_PIX_24b) 
		    {
		       pix=(unsigned int)byte[i++]<<16;
		       pix|=(unsigned int)byte[i++]<<8;
		       pix|=(unsigned int)byte[i++];
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    pix=(unsigned int)word[i++];
		  else
		    pix=(unsigned int)byte[i++];
		  decode_pixel(code,&p,pix);	
		  c->p[x][y].r=p.r;c->p[x][y].g=p.g;c->p[x][y].b=p.b;
	       }
	  }
	break;
      case FLAG_RES_4x4:
	for (y=0;y<4;y++)
	  {
	     for(x=0;x<4;x++)
	       {
		  if (code==FLAG_PIX_24b) 
		    {
		       pix=(unsigned int)byte[i++]<<16;
		       pix|=(unsigned int)byte[i++]<<8;
		       pix|=(unsigned int)byte[i++];
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    pix=(unsigned int)word[i++];
		  else
		    pix=(unsigned int)byte[i++];
		  decode_pixel(code,&p,pix);
		  for(yy=0;yy<2;yy++)
		    {
		       for(xx=0;xx<2;xx++)
			 {
			    c->p[(x<<1)+xx][(y<<1)+yy].r=p.r;
			    c->p[(x<<1)+xx][(y<<1)+yy].g=p.g;
			    c->p[(x<<1)+xx][(y<<1)+yy].b=p.b;
			 }
		    }
	       }
	  }
	break;
      case FLAG_RES_2x2:
	for (y=0;y<2;y++)
	  {
	     for(x=0;x<2;x++)
	       {
		  if (code==FLAG_PIX_24b) 
		    {
		       pix=(unsigned int)byte[i++]<<16;
		       pix|=(unsigned int)byte[i++]<<8;
		       pix|=(unsigned int)byte[i++];
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    pix=(unsigned int)word[i++];
		  else
		    pix=(unsigned int)byte[i++];
		  decode_pixel(code,&p,pix);
		  for(yy=0;yy<4;yy++)
		    {
		       for(xx=0;xx<4;xx++)
			 {
			    c->p[(x<<2)+xx][(y<<2)+yy].r=p.r;
			    c->p[(x<<2)+xx][(y<<2)+yy].g=p.g;
			    c->p[(x<<2)+xx][(y<<2)+yy].b=p.b;
			 }
		    }
	       }
	  }
	break;
      case FLAG_RES_1x1:
	for (y=0;y<1;y++)
	  {
	     for(x=0;x<1;x++)
	       {
		  if (code==FLAG_PIX_24b) 
		    {
		       pix=(unsigned int)byte[i++]<<16;
		       pix|=(unsigned int)byte[i++]<<8;
		       pix|=(unsigned int)byte[i++];
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    pix=(unsigned int)word[i++];
		  else
		    pix=(unsigned int)byte[i++];
		  decode_pixel(code,&p,pix);
		  for(yy=0;yy<8;yy++)
		    {
		       for(xx=0;xx<8;xx++)
			 {
			    c->p[(x<<3)+xx][(y<<3)+yy].r=p.r;
			    c->p[(x<<3)+xx][(y<<3)+yy].g=p.g;
			    c->p[(x<<3)+xx][(y<<3)+yy].b=p.b;
			 }
		    }
	       }
	  }
	break;
      default:
	break;
     }
}

RawChunk24 *load_chunk_chunks_m(unsigned char *m, int *size, int *w, int *h, int *x, int *y)
{
   RawChunk24 *c;
   int i,code,bpp,cd,sz;
   void *data;
   char s[10];
   unsigned char *ptr;
   int *ip;
   
   if (!m) 
     {
	*size=0;
	return NULL;
     }
   ptr=m+8;
   ip=(int *)ptr;
   *w=*ip;ip++;ptr+=sizeof(int);
   *h=*ip;ip++;ptr+=sizeof(int);
   *x=*ip;ip++;ptr+=sizeof(int);
   *y=*ip;ip++;ptr+=sizeof(int);

   c=(RawChunk24 *)malloc(sizeof(RawChunk24)**x**y);
   if (!c)
     {
	*size=ptr-m;
	return NULL;
     }
   for(i=0;i<*x**y;i++)
     {
	ip=(int *)ptr;
	code=*ip;ptr+=(sizeof(int));
	cd=code&FLAG_PIX_MSK;
	sz=code&FLAG_RES_MSK;
	if (cd==FLAG_PIX_24b) bpp=3;
	else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
		 (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
	else bpp=1;
	if (sz==FLAG_RES_8x8) bpp*=64;
	else if (sz==FLAG_RES_4x4) bpp*=16;
	else if (sz==FLAG_RES_2x2) bpp*=4;
	data=(void *)ptr;
	ptr+=bpp;
	decode_chunk(code,data,&c[i]);
     }
   *size=ptr-m;
   return c;
}

RawChunk24 *load_chunk_chunks_f(FILE *f, int *w, int *h, int *x, int *y)
{
   RawChunk24 *c;
   int i,code,bpp,cd,sz;
   void *data;
   char s[10];
   
   if (!f) return NULL;
   fread(s,1,8,f);
   s[8]=0;
   fread(w,sizeof(unsigned int),1,f);
   fread(h,sizeof(unsigned int),1,f);
   fread(x,sizeof(unsigned int),1,f);
   fread(y,sizeof(unsigned int),1,f);

   c=(RawChunk24 *)malloc(sizeof(RawChunk24)**x**y);
   if (!c)
     {
	return NULL;
     }
   data=(void *)malloc(64*3);
   if (!data)
     {
	free(c);
	return NULL;
     }
   for(i=0;i<*x**y;i++)
     {
	fread(&code,sizeof(unsigned int),1,f);
	cd=code&FLAG_PIX_MSK;
	sz=code&FLAG_RES_MSK;
	if (cd==FLAG_PIX_24b) bpp=3;
	else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
		 (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
	else bpp=1;
	if (sz==FLAG_RES_8x8) bpp*=64;
	else if (sz==FLAG_RES_4x4) bpp*=16;
	else if (sz==FLAG_RES_2x2) bpp*=4;
	fread(data,bpp,1,f);
	decode_chunk(code,data,&c[i]);
     }
   free(data);
   return c;
}

int load_chunks_frame_f(FILE *f, int *w, int *h, int *x, int *y, RawChunk24 *c)
{
   int i,code,bpp,cd,sz;
   void *data;

   data=(void *)malloc(64*3);
   if (!data)
     {
	return 0;
     }
   if (!fread(&code,sizeof(unsigned int),1,f))
     {
	return -1;
     }
   if (code==0xffffffff) 
     {
	free(data);
	return 0;
     }
   if (code==0xfffffffe) 
     {
	free(data);
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (sz==FLAG_RES_8x8) bpp*=64;
   else if (sz==FLAG_RES_4x4) bpp*=16;
   else if (sz==FLAG_RES_2x2) bpp*=4;
   if (!fread(data,bpp,1,f)) 
     {
	fprintf(stderr,"eof\n");
	return -1;
     }
   i=code>>6;
   decode_chunk(code,data,&c[i]);
   free(data);
   return 1;
}

RawChunk24 *load_chunks_chunks(char *file, int *w, int *h, int *x, int *y)
{
   RawChunk24 *c;
   FILE *f;
   int i,code,bpp,cd,sz;
   void *data;
   char s[10];
   
   f=fopen(file,"r");
   if (!f) return NULL;
   fread(s,1,8,f);
   s[8]=0;
   fread(w,sizeof(unsigned int),1,f);
   fread(h,sizeof(unsigned int),1,f);
   fread(x,sizeof(unsigned int),1,f);
   fread(y,sizeof(unsigned int),1,f);

   c=(RawChunk24 *)malloc(sizeof(RawChunk24)**x**y);
   if (!c)
     {
	fclose(f);
	return NULL;
     }
   data=(void *)malloc(64*3);
   if (!data)
     {
	free(c);
	fclose(f);
	return NULL;
     }
   for(i=0;i<*x**y;i++)
     {
	fread(&code,sizeof(unsigned int),1,f);
	cd=code&FLAG_PIX_MSK;
	sz=code&FLAG_RES_MSK;
	if (cd==FLAG_PIX_24b) bpp=3;
	else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
		 (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
	else bpp=1;
	if (sz==FLAG_RES_8x8) bpp*=64;
	else if (sz==FLAG_RES_4x4) bpp*=16;
	else if (sz==FLAG_RES_2x2) bpp*=4;
	fread(data,bpp,1,f);
	decode_chunk(code,data,&c[i]);
     }
   free(data);
   fclose(f);
   return c;
}

unsigned char *memopen(char *file, int *size)
{
   struct stat st;
   void *buf;
   int fd;

   stat(file,&st);
   buf=(void *)malloc(st.st_size);
   if (!buf) return NULL;
   fd=open(file,O_RDONLY);
   read(fd,buf,st.st_size);
   close(fd);
   return buf;
}

int main(int argc, char **argv)
{
   RawChunk24 *c,*c1,*c2;
   int w,h,x,y;
   int i,code,size,ii,j,sz;
   void *data;
   void *buf;
   unsigned int maxloss,maxdelta,loss;
   unsigned char *ppmdata;
   FILE *f,*ff;
   unsigned char *m,*mm;
   char s[1024];
   char *block;
   int loop;
   int mem,pos;
   int glob,root;
   int cnt;
   struct timeval tv;
   double t1,t2,tt,tot;
   struct sigaction sa;
   char file[1024];
   
   shmp=0;
   glob=2;
   mem=1;
   loop=1;
   root=0;
   sa.sa_handler = qit;
   sa.sa_flags = 0;
   sigemptyset (&sa.sa_mask);
   sigaction(SIGHUP,&sa,(struct sigaction *)0);
   sigaction(SIGINT,&sa,(struct sigaction *)0);
   sigaction(SIGQUIT,&sa,(struct sigaction *)0);
   sigaction(SIGTERM,&sa,(struct sigaction *)0);
   sigaction(SIGABRT,&sa,(struct sigaction *)0);
   sigaction(SIGPIPE,&sa,(struct sigaction *)0);
   
   if (argc<2)
     {
	printf("usage:\n\t%s anim_file [-file] [-noloop] [-glob n] [-root] [-shmp]\n",argv[0]);
	printf("\t-file : Play file off disk rather than memory\n");
	printf("\t-noloop : Do not loop animation (play once then exit)\n");
	printf("\t-glob n: Set display update globsize to n\n");
	printf("\t-root : Play animation in root window\n");
	printf("\t-shmp : Use shared pixmaps instead of shared images\n");
	exit(0);
     }
   if (argc>1)
     {
	if (!strcmp("-h",argv[1]))
	  {
	     printf("usage:\n\t%s anim_file [-file] [-noloop] [-glob n] [-root] [-shmp]\n",argv[0]);
	     printf("\t-file : Play file off disk rather than memory\n");
	     printf("\t-noloop : Do not loop animation (play once then exit)\n");
	     printf("\t-glob n: Set display update globsize to n\n");
	     printf("\t-root : Play animation in root window\n");
	     printf("\t-shmp : Use shared pixmaps instead of shared images\n");
	     exit(0);
	  }
     }
   if (argc>1)
     {
	strcpy(file,argv[1]);
	for(i=2;i<argc;i++)
	  {
	     if (!strcmp("-file",argv[i])) mem=0;
	     else if (!strcmp("-noloop",argv[i])) loop=0;
	     else if (!strcmp("-root",argv[i])) root=1;
	     else if (!strcmp("-shmp",argv[i])) shmp=1;
	     else if (!strcmp("-glob",argv[i++])) glob=atoi(argv[i]);
	     else if (!strcmp("-h",argv[i]))
	       {
		  printf("usage:\n\t%s anim_file [-file] [-noloop] [-glob n]\n");
		  printf("\t-file : Play file off disk rather than memory\n");
		  printf("\t-noloop : Do not loop animation (play once then exit)\n");
		  printf("\t-glob n: Set display update globsize to n\n");
		  printf("\t-root n: Play animation in root window\n");
		  exit(0);
	       }
	  }
     }
   if (argc>=0)
     {
	if (mem) 
	  {
	     m=memopen(file,&size);
	     mm=m;
	     c=load_chunk_chunks_m(mm,&sz,&w,&h,&x,&y);
	     mm+=sz;
	  }
	else
	  {
	     f=fopen(file,"r");
	     c=load_chunk_chunks_f(f,&w,&h,&x,&y);
	  }
	block=(char *)malloc(x*y);
	X_OpenWindow(0,0,w,h,root);X_SetupImage(w,h);X_Sync();
	X_Dechunkify(w,h,x,y,c);free(c);X_Sync();
	j=0;cnt=0;
	for (i=0;i<x*y;i++) block[i]=0;
	if (mem)
	  {
	     t1=0;
	     while ((ii=X_Load_chunks_frame_m(mm,&sz,&w,&h,&x,&y,c))>-3)
	       {
		  mm+=sz;
		  if (ii>0) block[ii-1]=1;
		  else if (ii<=0)
		    {
		       j++;
		       X_Sync();
		       cnt+=X_DisplayImage(block,x,y,glob);
		       for (i=0;i<x*y;i++) block[i]=0;
		       if (ii<0)
			 {
			    gettimeofday(&tv,NULL);
			    t2=(double)tv.tv_sec+((double)tv.tv_usec/1000000);
			    tt=t2-t1;
			    printf("frames %i fps: %2.2f deltaavg: %3.2f%%\r",j,(double)j/tt,(double)cnt*100/(double)(j*x*y));
			    fflush(stdout);
			 }
		    }
		  if ((ii==-1)&&(!loop)) break;
		  else if (ii==-2)
		    {
		       mm=m+24;
		       gettimeofday(&tv,NULL);
		       t1=(double)tv.tv_sec+((double)tv.tv_usec/1000000);
		       while (X_Skip_chunks_frame_m(mm,&sz,&w,&h,&x,&y,c)>0)
			 mm+=sz;
		       mm+=sz;
		       j=1;cnt=0;
		    }
	       }
	  }
	else
	  {
	     t1=0;
	     while ((ii=X_Load_chunks_frame_f(f,&w,&h,&x,&y,c))>-3)
	       {
		  if (ii>0) block[ii-1]=1;
		  else if (ii<=0)
		    {
		       j++;
		       X_Sync();
		       cnt+=X_DisplayImage(block,x,y,glob);
		       for (i=0;i<x*y;i++) block[i]=0;
		       if (ii<0)
			 {
			    gettimeofday(&tv,NULL);
			    t2=(double)tv.tv_sec+((double)tv.tv_usec/1000000);
			    tt=t2-t1;
			    printf("frames %i fps: %2.2f deltaavg: %3.2f%%\r",j,(double)j/tt,(double)cnt*100/(double)(j*x*y));
			    fflush(stdout);
			 }
		    }
		  if ((ii==-1)&&(!loop)) break;
		  else if ((ii==-2)&&(loop))
		    {
		       fseek(f,SEEK_SET,0);
		       fread(s,1,8,f);
		       fread(&w,sizeof(unsigned int),1,f);
		       fread(&h,sizeof(unsigned int),1,f);
		       fread(&x,sizeof(unsigned int),1,f);
		       fread(&y,sizeof(unsigned int),1,f);
		       gettimeofday(&tv,NULL);
		       t1=(double)tv.tv_sec+((double)tv.tv_usec/1000000);
		       while (X_Skip_chunks_frame_f(f,&sz,&w,&h,&x,&y,c)>0);
		       j=1;cnt=0;
		    }
	       }
	     fclose(f);
	  }
     }
}
