#include "eanim.h"

unsigned char *dechunkify(int w, int h, int x, int y, RawChunk24 *c)
{
   int xx,yy,x2,y2,i,j;
   unsigned char *data;
   
   data=(unsigned char *)malloc(w*h*3);
   if (!data) return NULL;
   i=0;j=0;
   for(yy=0;yy<y;yy++)
     {
	for(y2=0;y2<8;y2++)
	  {
	     for(xx=0;xx<x;xx++)
	       {
		  i=(yy*x)+xx;
		  for(x2=0;x2<8;x2++)
		    {
		       data[j++]=c[i].p[x2][y2].r;
		       data[j++]=c[i].p[x2][y2].g;
		       data[j++]=c[i].p[x2][y2].b;
		    }
	       }
	  }
     }
   return data;
}

RawChunk24 *chunkify(int w, int h, unsigned char *data, int *x, int *y)
{
   int xx,yy,i,cnum;
   RawChunk24 *c;
   
   *x=w>>3;if (w&0x7) *x++;
   *y=h>>3;if (h&0x7) *y++;
   c=(RawChunk24 *)malloc(sizeof(RawChunk24)**x**y);
   if (!c) return NULL;
   i=0;
   if (data)
     {
	for (yy=0;yy<h;yy++)
	  {
	     for (xx=0;xx<w;xx++)
	       {
		  cnum=((yy>>3)**x)+(xx>>3);
		  c[cnum].p[xx&7][yy&7].r=data[i++];
		  c[cnum].p[xx&7][yy&7].g=data[i++];
		  c[cnum].p[xx&7][yy&7].b=data[i++];
	       }
	  }
     }
   else
     memset(c,0,sizeof(RawChunk24)**x**y);
   return c;
}

void decode_pixel(int code, Pixel24 *p, int pix)
{
   unsigned int r,g,b;
   
   switch (code)
     {
      case FLAG_PIX_24b:
	r=(pix>>16)&0xff;
	g=(pix>>8)&0xff;
	b=(pix)&0xff;
	break;
      case FLAG_PIX_16b:
	r=(pix>>8)&0xf8;
	g=(pix>>3)&0xfc;
	b=(pix<<3)&0xf8;
	break;
      case FLAG_PIX_332:
	r=(pix)&0xe0;
	g=(pix<<3)&0xe0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_R:
	r=pix;
	g=0;
	b=0;
	break;
      case FLAG_PIX_G:
	r=0;
	g=pix;
	b=0;
	break;
      case FLAG_PIX_B:
	r=0;
	g=0;
	b=pix;
	break;
      case FLAG_PIX_RG:
	r=(pix>>8);
	g=pix&0xff;
	b=0;
	break;
      case FLAG_PIX_RB:
	r=(pix>>8);
	g=0;
	b=pix&0xff;
	break;
      case FLAG_PIX_GB:
	r=0;
	g=(pix>>8);
	b=pix&0xff;
	break;
      case FLAG_PIX_RG8:
	r=pix&0xf0;
	g=(pix<<4)&0xf0;
	b=0;
	break;
      case FLAG_PIX_RB8:
	r=pix&0xf0;
	g=0;
	b=(pix<<4)&0xf0;
	break;
      case FLAG_PIX_GB8:
	r=0;
	g=pix&0xf0;
	b=(pix<<4)&0xf0;
	break;
      case FLAG_PIX_GRY:
	r=pix;
	g=pix;
	b=pix;
	break;
      case FLAG_PIX_422:
	r=pix&0xf0;
	g=(pix<<4)&0xc0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_242:
	r=pix&0xc0;
	g=(pix<<2)&0xf0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_224:
	r=pix&0xc0;
	g=(pix<<2)&0xc0;
	b=(pix<<4)&0xf0;
	break;
      default:
	break;
     }
   p->r=(unsigned char)r;
   p->g=(unsigned char)g;
   p->b=(unsigned char)b;
}

int encode_pixel(int code, Pixel24 *p)
{
   int pix;
   
   pix=0;
   switch (code)
     {
      case FLAG_PIX_24b:
	pix=(((unsigned int)p->r<<16)&0xff0000)|
	  (((unsigned int)p->g<<8)&0x00ff00)|
	  (((unsigned int)p->b)&0x0000ff);
	break;
      case FLAG_PIX_16b:
	pix=(((unsigned int)p->r<<8)&0xf800)|
	  (((unsigned int)p->g<<3)&0x07e0)|
	  (((unsigned int)p->b>>3)&0x001f);
	break;
      case FLAG_PIX_332:
	pix=(((unsigned int)p->r)&0xe0)|
	  (((unsigned int)p->g>>3)&0x1c)|
	  (((unsigned int)p->b>>6)&0x03);
	break;
      case FLAG_PIX_R:
	pix=(unsigned int)p->r;
	break;
      case FLAG_PIX_G:
	pix=(unsigned int)p->g;
	break;
      case FLAG_PIX_B:
	pix=(unsigned int)p->b;
	break;
      case FLAG_PIX_RG:
	pix=(((unsigned int)p->r<<8)&0xff00)|
	  (((unsigned int)p->g)&0xff);
	break;
      case FLAG_PIX_RB:
	pix=(((unsigned int)p->r<<8)&0xff00)|
	  (((unsigned int)p->b)&0xff);
	break;
      case FLAG_PIX_GB:
	pix=(((unsigned int)p->g<<8)&0xff00)|
	  (((unsigned int)p->b)&0xff);
	break;
      case FLAG_PIX_RG8:
	pix=(((unsigned int)p->r)&0xf0)|
	  (((unsigned int)p->g>>4)&0x0f);
	break;
      case FLAG_PIX_RB8:
	pix=(((unsigned int)p->r)&0xf0)|
	  (((unsigned int)p->b>>4)&0x0f);
	break;
      case FLAG_PIX_GB8:
	pix=(((unsigned int)p->g)&0xf0)|
	  (((unsigned int)p->b>>4)&0x0f);
	break;
      case FLAG_PIX_GRY:
	pix=((((unsigned int)p->r))+
	  (((unsigned int)p->g))+
	  (((unsigned int)p->b)))/3;
	break;
      case FLAG_PIX_422:
	pix=(((unsigned int)p->r)&0xf0)|
	  (((unsigned int)p->g>>4)&0x0c)|
	  (((unsigned int)p->b>>6)&0x03);
	break;
      case FLAG_PIX_242:
	pix=(((unsigned int)p->r)&0xc0)|
	  (((unsigned int)p->g>>2)&0x3c)|
	  (((unsigned int)p->b>>6)&0x03);
	break;
      case FLAG_PIX_224:
	pix=(((unsigned int)p->r)&0xc0)|
	  (((unsigned int)p->g>>2)&0x30)|
	  (((unsigned int)p->b>>4)&0x0f);
	break;
      default:
	break;
     }
   return pix;
}

int encode_loss(int code, int pix, Pixel24 *p)
{
   unsigned int r,g,b;
   int loss;
   
   switch (code)
     {
      case FLAG_PIX_24b:
	r=(pix>>16)&0xff;
	g=(pix>>8)&0xff;
	b=(pix)&0xff;
	break;
      case FLAG_PIX_16b:
	r=(pix>>8)&0xf8;
	g=(pix>>3)&0xfc;
	b=(pix<<3)&0xf8;
	break;
      case FLAG_PIX_332:
	r=(pix)&0xe0;
	g=(pix<<3)&0xe0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_R:
	r=pix;
	g=0;
	b=0;
	break;
      case FLAG_PIX_G:
	r=0;
	g=pix;
	b=0;
	break;
      case FLAG_PIX_B:
	r=0;
	g=0;
	b=pix;
	break;
      case FLAG_PIX_RG:
	r=(pix>>8);
	g=pix&0xff;
	b=0;
	break;
      case FLAG_PIX_RB:
	r=(pix>>8);
	g=0;
	b=pix&0xff;
	break;
      case FLAG_PIX_GB:
	r=0;
	g=(pix>>8);
	b=pix&0xff;
	break;
      case FLAG_PIX_RG8:
	r=pix&0xf0;
	g=(pix<<4)&0xf0;
	b=0;
	break;
      case FLAG_PIX_RB8:
	r=pix&0xf0;
	g=0;
	b=(pix<<4)&0xf0;
	break;
      case FLAG_PIX_GB8:
	r=0;
	g=pix&0xf0;
	b=(pix<<4)&0xf0;
	break;
      case FLAG_PIX_GRY:
	r=pix;
	g=pix;
	b=pix;
	break;
      case FLAG_PIX_422:
	r=pix&0xf0;
	g=(pix<<4)&0xc0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_242:
	r=pix&0xc0;
	g=(pix<<2)&0xf0;
	b=(pix<<6)&0xc0;
	break;
      case FLAG_PIX_224:
	r=pix&0xc0;
	g=(pix<<2)&0xc0;
	b=(pix<<4)&0xf0;
	break;
      default:
	break;
     }
   if (p->r>r) r=p->r-r;
   else r=r-p->r;
   if (p->g>g) g=p->g-g;
   else g=g-p->g;
   if (p->b>b) b=p->b-b;
   else b=b-p->b;
   loss=r+g+b;
   return loss;
}

void decode_chunk(int code_type, void *data, RawChunk24 *c)
{
   Pixel24 p;
   int bpp,x,y,xx,yy,res,i,code;
   unsigned int pix;
   unsigned char *byte;
   unsigned short *word;
   
   res=code_type&FLAG_RES_MSK;
   code=code_type&FLAG_PIX_MSK;
   if (code==FLAG_PIX_24b) bpp=3;
   else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
	    (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (bpp==2) word=(unsigned short *)data;
   else byte=(unsigned char *)data;
   i=0;
   switch(res)
     {
      case FLAG_RES_8x8:
	for (y=0;y<8;y++)
	  {
	     for(x=0;x<8;x++)
	       {
		  if (code==FLAG_PIX_24b) 
		    {
		       pix=(unsigned int)byte[i++]<<16;
		       pix|=(unsigned int)byte[i++]<<8;
		       pix|=(unsigned int)byte[i++];
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    pix=(unsigned int)word[i++];
		  else
		    pix=(unsigned int)byte[i++];
		  decode_pixel(code,&p,pix);	
		  c->p[x][y].r=p.r;c->p[x][y].g=p.g;c->p[x][y].b=p.b;
	       }
	  }
	break;
      case FLAG_RES_4x4:
	for (y=0;y<4;y++)
	  {
	     for(x=0;x<4;x++)
	       {
		  if (code==FLAG_PIX_24b) 
		    {
		       pix=(unsigned int)byte[i++]<<16;
		       pix|=(unsigned int)byte[i++]<<8;
		       pix|=(unsigned int)byte[i++];
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    pix=(unsigned int)word[i++];
		  else
		    pix=(unsigned int)byte[i++];
		  decode_pixel(code,&p,pix);
		  for(yy=0;yy<2;yy++)
		    {
		       for(xx=0;xx<2;xx++)
			 {
			    c->p[(x<<1)+xx][(y<<1)+yy].r=p.r;
			    c->p[(x<<1)+xx][(y<<1)+yy].g=p.g;
			    c->p[(x<<1)+xx][(y<<1)+yy].b=p.b;
			 }
		    }
	       }
	  }
	break;
      case FLAG_RES_2x2:
	for (y=0;y<2;y++)
	  {
	     for(x=0;x<2;x++)
	       {
		  if (code==FLAG_PIX_24b) 
		    {
		       pix=(unsigned int)byte[i++]<<16;
		       pix|=(unsigned int)byte[i++]<<8;
		       pix|=(unsigned int)byte[i++];
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    pix=(unsigned int)word[i++];
		  else
		    pix=(unsigned int)byte[i++];
		  decode_pixel(code,&p,pix);
		  for(yy=0;yy<4;yy++)
		    {
		       for(xx=0;xx<4;xx++)
			 {
			    c->p[(x<<2)+xx][(y<<2)+yy].r=p.r;
			    c->p[(x<<2)+xx][(y<<2)+yy].g=p.g;
			    c->p[(x<<2)+xx][(y<<2)+yy].b=p.b;
			 }
		    }
	       }
	  }
	break;
      case FLAG_RES_1x1:
	for (y=0;y<1;y++)
	  {
	     for(x=0;x<1;x++)
	       {
		  if (code==FLAG_PIX_24b) 
		    {
		       pix=(unsigned int)byte[i++]<<16;
		       pix|=(unsigned int)byte[i++]<<8;
		       pix|=(unsigned int)byte[i++];
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    pix=(unsigned int)word[i++];
		  else
		    pix=(unsigned int)byte[i++];
		  decode_pixel(code,&p,pix);
		  for(yy=0;yy<8;yy++)
		    {
		       for(xx=0;xx<8;xx++)
			 {
			    c->p[(x<<3)+xx][(y<<3)+yy].r=p.r;
			    c->p[(x<<3)+xx][(y<<3)+yy].g=p.g;
			    c->p[(x<<3)+xx][(y<<3)+yy].b=p.b;
			 }
		    }
	       }
	  }
	break;
      default:
	break;
     }
}

void *encode_chunk(int code_type, unsigned int *loss, RawChunk24 *c)
{
   Pixel24 p;
   int res,code,pix,i,x,y,xx,yy;
   unsigned char *byte;
   unsigned short *word;
   void *ptr;
   unsigned int diff;
   unsigned int r,g,b;
   
   res=code_type&FLAG_RES_MSK;
   code=code_type&FLAG_PIX_MSK;
   i=0;diff=0;
   switch(res)
     {
      case FLAG_RES_8x8:
	if (code==FLAG_PIX_24b) 
	  {
	     ptr=(void *)malloc(64*3);
	     byte=(unsigned char *)ptr;
	  }
	else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
		 (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
	  {
	     ptr=(void *)malloc(64*2);
	     word=(unsigned short *)ptr;
	  }
	else
	  {
	     ptr=(void *)malloc(64*1);
	     byte=(unsigned char *)ptr;
	  }
	for (y=0;y<8;y++)
	  {
	     for(x=0;x<8;x++)
	       {
		  p.r=c->p[x][y].r;p.g=c->p[x][y].g;p.b=c->p[x][y].b;
		  pix=encode_pixel(code,&p);
		  diff+=encode_loss(code,pix,&p);
		  if (code==FLAG_PIX_24b) 
		    {
		       byte[i++]=(unsigned char)((pix>>16)&0xff);
		       byte[i++]=(unsigned char)((pix>>8)&0xff);
		       byte[i++]=(unsigned char)((pix)&0xff);
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    word[i++]=(unsigned short)pix;
		  else
		    byte[i++]=(unsigned char)pix;
	       }
	  }
	break;
      case FLAG_RES_4x4:
	if (code==FLAG_PIX_24b) 
	  {
	     ptr=(void *)malloc(16*3);
	     byte=(unsigned char *)ptr;
	  }
	else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
		 (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
	  {
	     ptr=(void *)malloc(16*2);
	     word=(unsigned short *)ptr;
	  }
	else
	  {
	     ptr=(void *)malloc(16*1);
	     byte=(unsigned char *)ptr;
	  }
	for (y=0;y<4;y++)
	  {
	     for(x=0;x<4;x++)
	       {
		  r=g=b=0;
		  for (yy=0;yy<2;yy++)
		    {
		       for (xx=0;xx<2;xx++)
			 {
			    r+=c->p[(x<<1)+xx][(y<<1)+yy].r;
			    g+=c->p[(x<<1)+xx][(y<<1)+yy].g;
			    b+=c->p[(x<<1)+xx][(y<<1)+yy].b;
			 }
		    }
		  p.r=r>>2;p.g=g>>2;p.b=b>>2;
		  pix=encode_pixel(code,&p);
		  for (yy=0;yy<2;yy++)
		    {
		       for (xx=0;xx<2;xx++)
			 {
			    diff+=encode_loss(code,pix,&c->p[(x<<1)+xx][(y<<1)+yy]);
			 }
		    }
		  if (code==FLAG_PIX_24b) 
		    {
		       byte[i++]=(unsigned char)((pix>>16)&0xff);
		       byte[i++]=(unsigned char)((pix>>8)&0xff);
		       byte[i++]=(unsigned char)((pix)&0xff);
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    word[i++]=(unsigned short)pix;
		  else
		    byte[i++]=(unsigned char)pix;
	       }
	  }
	break;
      case FLAG_RES_2x2:
	if (code==FLAG_PIX_24b) 
	  {
	     ptr=(void *)malloc(4*3);
	     byte=(unsigned char *)ptr;
	  }
	else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
		 (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
	  {
	     ptr=(void *)malloc(4*2);
	     word=(unsigned short *)ptr;
	  }
	else
	  {
	     ptr=(void *)malloc(4*1);
	     byte=(unsigned char *)ptr;
	  }
	for (y=0;y<2;y++)
	  {
	     for(x=0;x<2;x++)
	       {
		  r=g=b=0;
		  for (yy=0;yy<4;yy++)
		    {
		       for (xx=0;xx<4;xx++)
			 {
			    r+=c->p[(x<<2)+xx][(y<<2)+yy].r;
			    g+=c->p[(x<<2)+xx][(y<<2)+yy].g;
			    b+=c->p[(x<<2)+xx][(y<<2)+yy].b;
			 }
		    }
		  p.r=r>>4;p.g=g>>4;p.b=b>>4;
		  pix=encode_pixel(code,&p);
		  for (yy=0;yy<4;yy++)
		    {
		       for (xx=0;xx<4;xx++)
			 {
			    diff+=encode_loss(code,pix,&c->p[(x<<2)+xx][(y<<2)+yy]);
			 }
		    }
		  if (code==FLAG_PIX_24b) 
		    {
		       byte[i++]=(unsigned char)((pix>>16)&0xff);
		       byte[i++]=(unsigned char)((pix>>8)&0xff);
		       byte[i++]=(unsigned char)((pix)&0xff);
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    word[i++]=(unsigned short)pix;
		  else
		    byte[i++]=(unsigned char)pix;
	       }
	  }
	break;
      case FLAG_RES_1x1:
	if (code==FLAG_PIX_24b) 
	  {
	     ptr=(void *)malloc(1*3);
	     byte=(unsigned char *)ptr;
	  }
	else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
		 (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
	  {
	     ptr=(void *)malloc(1*2);
	     word=(unsigned short *)ptr;
	  }
	else
	  {
	     ptr=(void *)malloc(1*1);
	     byte=(unsigned char *)ptr;
	  }
	for (y=0;y<1;y++)
	  {
	     for(x=0;x<1;x++)
	       {
		  r=g=b=0;
		  for (yy=0;yy<8;yy++)
		    {
		       for (xx=0;xx<8;xx++)
			 {
			    r+=c->p[(x<<3)+xx][(y<<3)+yy].r;
			    g+=c->p[(x<<3)+xx][(y<<3)+yy].g;
			    b+=c->p[(x<<3)+xx][(y<<3)+yy].b;
			 }
		    }
		  p.r=r>>6;p.g=g>>6;p.b=b>>6;
		  pix=encode_pixel(code,&p);
		  for (yy=0;yy<8;yy++)
		    {
		       for (xx=0;xx<8;xx++)
			 {
			    diff+=encode_loss(code,pix,&c->p[(x<<3)+xx][(y<<3)+yy]);
			 }
		    }
		  if (code==FLAG_PIX_24b) 
		    {
		       byte[i++]=(unsigned char)((pix>>16)&0xff);
		       byte[i++]=(unsigned char)((pix>>8)&0xff);
		       byte[i++]=(unsigned char)((pix)&0xff);
		    }
		  else if ((code==FLAG_PIX_16b)||(code==FLAG_PIX_RG)||
			   (code==FLAG_PIX_RB)||(code==FLAG_PIX_GB))
		    word[i++]=(unsigned short)pix;
		  else
		    byte[i++]=(unsigned char)pix;
	       }
	  }
	break;
      default:
	break;
     }
   *loss=diff;
   return ptr;
}

RawChunk24 *load_chunk_chunks_f(FILE *f, int *w, int *h, int *x, int *y)
{
   RawChunk24 *c;
   int i,code,bpp,cd,sz;
   void *data;
   char s[10];
   
   if (!f) return NULL;
   fread(s,1,8,f);
   s[8]=0;
   fread(w,sizeof(unsigned int),1,f);
   fread(h,sizeof(unsigned int),1,f);
   fread(x,sizeof(unsigned int),1,f);
   fread(y,sizeof(unsigned int),1,f);

   c=(RawChunk24 *)malloc(sizeof(RawChunk24)**x**y);
   if (!c)
     {
	return NULL;
     }
   data=(void *)malloc(64*3);
   if (!data)
     {
	free(c);
	return NULL;
     }
   for(i=0;i<*x**y;i++)
     {
	fread(&code,sizeof(unsigned int),1,f);
	cd=code&FLAG_PIX_MSK;
	sz=code&FLAG_RES_MSK;
	if (cd==FLAG_PIX_24b) bpp=3;
	else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
		 (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
	else bpp=1;
	if (sz==FLAG_RES_8x8) bpp*=64;
	else if (sz==FLAG_RES_4x4) bpp*=16;
	else if (sz==FLAG_RES_2x2) bpp*=4;
	fread(data,bpp,1,f);
	decode_chunk(code,data,&c[i]);
     }
   free(data);
   return c;
}

int load_chunks_frame_f(FILE *f, int *w, int *h, int *x, int *y, RawChunk24 *c)
{
   int i,code,bpp,cd,sz;
   void *data;

   data=(void *)malloc(64*3);
   if (!data)
     {
	return 0;
     }
   if (!fread(&code,sizeof(unsigned int),1,f))
     {
	return -1;
     }
   if (code==0xffffffff) 
     {
	free(data);
	return 0;
     }
   if (code==0xfffffffe) 
     {
	free(data);
	return -2;
     }
   cd=code&FLAG_PIX_MSK;
   sz=code&FLAG_RES_MSK;
   if (cd==FLAG_PIX_24b) bpp=3;
   else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
	    (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
   else bpp=1;
   if (sz==FLAG_RES_8x8) bpp*=64;
   else if (sz==FLAG_RES_4x4) bpp*=16;
   else if (sz==FLAG_RES_2x2) bpp*=4;
   if (!fread(data,bpp,1,f)) 
     {
	fprintf(stderr,"eof\n");
	return -1;
     }
   i=code>>6;
   decode_chunk(code,data,&c[i]);
   free(data);
   return 1;
}

RawChunk24 *load_chunks_chunks(char *file, int *w, int *h, int *x, int *y)
{
   RawChunk24 *c;
   FILE *f;
   int i,code,bpp,cd,sz;
   void *data;
   char s[10];
   
   f=fopen(file,"r");
   if (!f) return NULL;
   fread(s,1,8,f);
   s[8]=0;
   fread(w,sizeof(unsigned int),1,f);
   fread(h,sizeof(unsigned int),1,f);
   fread(x,sizeof(unsigned int),1,f);
   fread(y,sizeof(unsigned int),1,f);

   c=(RawChunk24 *)malloc(sizeof(RawChunk24)**x**y);
   if (!c)
     {
	fclose(f);
	return NULL;
     }
   data=(void *)malloc(64*3);
   if (!data)
     {
	free(c);
	fclose(f);
	return NULL;
     }
   for(i=0;i<*x**y;i++)
     {
	fread(&code,sizeof(unsigned int),1,f);
	cd=code&FLAG_PIX_MSK;
	sz=code&FLAG_RES_MSK;
	if (cd==FLAG_PIX_24b) bpp=3;
	else if ((cd==FLAG_PIX_16b)||(cd==FLAG_PIX_RG)||
		 (cd==FLAG_PIX_RB)||(cd==FLAG_PIX_GB)) bpp=2;
	else bpp=1;
	if (sz==FLAG_RES_8x8) bpp*=64;
	else if (sz==FLAG_RES_4x4) bpp*=16;
	else if (sz==FLAG_RES_2x2) bpp*=4;
	fread(data,bpp,1,f);
	decode_chunk(code,data,&c[i]);
     }
   free(data);
   fclose(f);
   return c;
}

RawChunk24 *load_ppm_chunks(char *file, int *w, int *h, int *x, int *y)
{
   unsigned char *data;
   FILE *f;
   char s[1024];
   char ss[1024];
   RawChunk24 *c;
   
   sprintf(s,"convert %s pnm:-",file);
   f=popen(s,"r");
   if (!f) return NULL;
   fgets(s,1024,f);
   sscanf(s,"%s",ss);
   if (strcmp("P6",ss)) 
     {
	pclose(f);
	return NULL;
     }
   fgets(s,1024,f);
   while (s[0]=='#') fgets(s,1024,f);
   sleep(1);
   sscanf(s,"%i %i",w,h);
   fgets(s,1024,f);
   data=(unsigned char *)malloc(*w**h*3);
   if (!data)
     {
	pclose(f);
	return NULL;
     }
   fread(data,1,*w**h*3,f);
   pclose(f);
   c=chunkify(*w,*h,data,x,y);
   free(data);
   return c;
}

void *get_best_chunk(RawChunk24 *cc, int *code, int maxloss, int *size)
{
   struct _chunk
     {
	int size;
	int loss;
	int code;
	void *data;
     } ch[64];
   int sz,cd,i,siz,cod;
   void *data;
   int sorted;
   int a,b,c;
   void *d;
   
   sz=0;cd=0;
   for (i=0;i<64;i++)
     {
	cod=cd<<2;
	ch[i].data=encode_chunk(cod|sz,&ch[i].loss,cc);
	ch[i].code=cod|sz;
	if (cod==FLAG_PIX_24b) siz=3;
	else if 
	  ((cod==FLAG_PIX_16b)||
	   (cod==FLAG_PIX_RG)||
	   (cod==FLAG_PIX_RB)||
	   (cod==FLAG_PIX_GB)) siz=2;
	else siz=1;
	if (sz==FLAG_RES_8x8) siz*=64;
	else if (sz==FLAG_RES_4x4) siz*=16;
	else if (sz==FLAG_RES_2x2) siz*=4;
	ch[i].size=siz;
	cd++;
	if(cd>15) 
	  {
	     cd=0;
	     sz++;
	  }
     }
   sorted=0;
   while(!sorted)
     {
	sorted=1;
	for (i=0;i<63;i++)
	  {
	     if (ch[i].size==ch[i+1].size)
	       {
		  if (ch[i].loss>ch[i+1].loss)
		    {
		       sorted=0;
		       a=ch[i].size;
		       b=ch[i].loss;
		       c=ch[i].code;
		       d=ch[i].data;
		       ch[i].size=ch[i+1].size;
		       ch[i].loss=ch[i+1].loss;
		       ch[i].code=ch[i+1].code;
		       ch[i].data=ch[i+1].data;
		       ch[i+1].size=a;
		       ch[i+1].loss=b;
		       ch[i+1].code=c;
		       ch[i+1].data=d;
		    }
	       }
	  }
     }

   for (i=63;i>=0;i--)
     {
	if (ch[i].loss<=maxloss)
	  {
	     data=ch[i].data;
	     ch[i].data=NULL;
	     *code=ch[i].code;
	     *size=ch[i].size;
	     i=-1;
	  }
     }
   for (i=0;i<64;i++)
     {
	if (ch[i].data) free(ch[i].data);
     }
   return data;
}

void *gendelta(RawChunk24 *c1, RawChunk24 *c2, int *code, int *size, 
	       int maxloss, int maxdelta, int *loss)
{
   void *d;
   int cd;
   int x,y,diff;
   
   d=get_best_chunk(c2,&cd,maxloss,size);
   if (!d) return NULL;
   decode_chunk(cd,d,c2);
   diff=0;
   for (y=0;y<8;y++)
     {
	for(x=0;x<8;x++)
	  {
	     if (c2->p[x][y].r<c1->p[x][y].r) diff+=c1->p[x][y].r-c2->p[x][y].r;
	     else diff+=c2->p[x][y].r-c1->p[x][y].r;
	     if (c2->p[x][y].g<c1->p[x][y].g) diff+=c1->p[x][y].g-c2->p[x][y].g;
	     else diff+=c2->p[x][y].g-c1->p[x][y].g;
	     if (c2->p[x][y].b<c1->p[x][y].b) diff+=c1->p[x][y].b-c2->p[x][y].b;
	     else diff+=c2->p[x][y].b-c1->p[x][y].b;
	  }
     }
   if (diff>=maxdelta) 
     {
	*loss=diff;
	*code=cd;
	return d;
     }
   else
     {
	*loss=diff;
	free(d);
	return NULL;
     }
}

int main(int argc, char **argv)
{
   RawChunk24 *c,*c1,*c2;
   int w,h,x,y;
   int i,code,size,ii;
   void *data;
   unsigned int maxloss,maxdelta,loss;
   unsigned char *ppmdata;
   FILE *f,*ff;
   char s[1024];
   int *curloss;
   int loop;
   
   loop=1;
   if (argc>3)
     {
	maxloss=atoi(argv[1])*48;
	maxdelta=atoi(argv[2])*48;
	c=load_ppm_chunks(argv[3],&w,&h,&x,&y);
	fwrite("EANIM-01",1,8,stdout);
	fwrite(&w,sizeof(unsigned int),1,stdout);
	fwrite(&h,sizeof(unsigned int),1,stdout);
	fwrite(&x,sizeof(unsigned int),1,stdout);
	fwrite(&y,sizeof(unsigned int),1,stdout);
	curloss=(int *)malloc(sizeof(int)*x*y);
	for (i=0;i<x*y;i++)
	  {
	     curloss[i]=0;
	     if (!(i%25)) 
	       {
		  fprintf(stderr,"0: %i/%i        \r",i,x*y);
		  fflush(stderr);
	       }
	     code=0;
	     data=get_best_chunk(&c[i],&code,maxloss,&size);
	     decode_chunk(code,data,&c[i]);
	     code|=i<<6;
	     fwrite(&code,sizeof(unsigned int),1,stdout);
	     fwrite(data,size,1,stdout);
	     free(data);
	  }
	c1=c;
	c2=NULL;
	for(ii=4;ii<argc;ii++)
	  {
	     c2=load_ppm_chunks(argv[ii],&w,&h,&x,&y);
	     for(i=0;i<x*y;i++)
	       {
		  code=0;
		  data=gendelta(&c1[i],&c2[i],&code,&size,maxloss,maxdelta,&loss);
		  if (!data) curloss[i]+=loss;
		  else curloss[i]=0;
		  if (curloss[i]>maxloss) curloss[i]=maxloss;
		  if (data)
		    {
		       code|=i<<6;
		       fwrite(&code,sizeof(unsigned int),1,stdout);
		       fwrite(data,size,1,stdout);
		       free(data);
		    }
		  if (!(i%25)) 
		    {
		       fprintf(stderr,"%i: %i/%i        \r",ii-3,i,x*y);
		       fflush(stderr);
		    }
	       }
	     if ((ii==argc-1)&&(loop))
	       code=0xfffffffe;
	     else
	       code=0xffffffff;
	     fwrite(&code,sizeof(unsigned int),1,stdout);
	     c=c1;c1=c2;c2=c;
	     free(c2);
	  }
     }
   else if (argc==2)
     {
	c=load_chunks_chunks(argv[1],&w,&h,&x,&y);
	ppmdata=dechunkify(w,h,x,y,c);
	printf("P6\n%i %i\n255\n",w,h);
	fwrite(ppmdata,(w*h*3),1,stdout);
     }
   else if (argc==3)
     {
	f=fopen(argv[1],"r");
	c=load_chunk_chunks_f(f,&w,&h,&x,&y);
	i=1;
	if (i<10)
	  sprintf(s,"%s000%i.ppm",argv[2],i);
	else if (i<100)
	  sprintf(s,"%s00%i.ppm",argv[2],i);
	else if (i<1000)
	  sprintf(s,"%s0%i.ppm",argv[2],i);
	else
	  sprintf(s,"%s%i.ppm",argv[2],i);
	ff=fopen(s,"w");
	ppmdata=dechunkify(w,h,x,y,c);
	fprintf(ff,"P6\n%i %i\n255\n",w,h);
	fwrite(ppmdata,(w*h*3),1,ff);
	fclose(ff);
	free(ppmdata);
	i++;
	while ((ii=load_chunks_frame_f(f,&w,&h,&x,&y,c))>=0)
	  {
	     fprintf(stderr,"%i:   \r",i);
	     fflush(stderr);
	     if (ii<=0)
	       {
		  if (i<10)
		    sprintf(s,"%s000%i.ppm",argv[2],i);
		  else if (i<100)
		    sprintf(s,"%s00%i.ppm",argv[2],i);
		  else if (i<1000)
		    sprintf(s,"%s0%i.ppm",argv[2],i);
		  else
		    sprintf(s,"%s%i.ppm",argv[2],i);
		  ff=fopen(s,"w");
		  ppmdata=dechunkify(w,h,x,y,c);
		  fprintf(ff,"P6\n%i %i\n255\n",w,h);
		  fwrite(ppmdata,(w*h*3),1,ff);
		  fclose(ff);
		  free(ppmdata);
		  i++;
	       }
	     if (ii==-1) break;
	     else if (ii==-2)
	       {
		  /*loop*/
	       }
	  }
	free(c);
	fclose(f);
     }
}
